/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/*
 *    Brutus e-cal backend factory class.
 *
 *    Authors:
 *        Jules Colding <colding@omesc.com>
 *
 *    Copyright (C) 2006-2007 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *    MA 02111-1307 USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib.h>
#include <glib/gstdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>

#include <server/brutus.h>
#include <session/brutusd.h>

#include "e-cal-backend-brutus-factory.h"
#include "e-cal-backend-brutus.h"

typedef struct {
	ECalBackendFactory parent_object;
} ECalBackendBrutusFactory;

typedef struct {
	ECalBackendBrutusClass parent_class;
} ECalBackendBrutusFactoryClass;

static void
e_cal_backend_brutus_factory_instance_init(ECalBackendBrutusFactory *factory)
{
	d("");
}

static const char*
_get_protocol(ECalBackendFactory *factory)
{
	d("");
	return "brutus";
}

static ECalBackend*
_todos_new_backend(ECalBackendFactory *factory,
		   ESource *source)
{
	d("");
	return g_object_new(e_cal_backend_brutus_get_type(),
			    "source",
			    source,
			    "kind",
			    ICAL_VTODO_COMPONENT,
			    NULL);
}

static icalcomponent_kind
_todos_get_kind(ECalBackendFactory *factory)
{
	d("");
	return ICAL_VTODO_COMPONENT;
}

static ECalBackend*
_events_new_backend(ECalBackendFactory *factory,
		    ESource *source)
{
	d("");
	return g_object_new(e_cal_backend_brutus_get_type(),
			    "source", 
			    source,
			    "kind",
			    ICAL_VEVENT_COMPONENT,
			    NULL);
}

static icalcomponent_kind
_events_get_kind(ECalBackendFactory *factory)
{
	d("");
	return ICAL_VEVENT_COMPONENT;
}

static void
todos_backend_factory_class_init(ECalBackendBrutusFactoryClass *klass)
{
	d("");
	E_CAL_BACKEND_FACTORY_CLASS(klass)->get_protocol = _get_protocol;
	E_CAL_BACKEND_FACTORY_CLASS(klass)->get_kind = _todos_get_kind;
	E_CAL_BACKEND_FACTORY_CLASS(klass)->new_backend  = _todos_new_backend;
}

static void
events_backend_factory_class_init(ECalBackendBrutusFactoryClass *klass)
{
	d("");
	E_CAL_BACKEND_FACTORY_CLASS(klass)->get_protocol = _get_protocol;
	E_CAL_BACKEND_FACTORY_CLASS(klass)->get_kind = _events_get_kind;
	E_CAL_BACKEND_FACTORY_CLASS(klass)->new_backend  = _events_new_backend;
}

static GType
events_backend_factory_get_type(GTypeModule *module)
{
	GType type;

	GTypeInfo info = {
		sizeof(ECalBackendBrutusFactoryClass),
		NULL, /* base_class_init */
		NULL, /* base_class_finalize */
		(GClassInitFunc)events_backend_factory_class_init,
		NULL, /* class_finalize */
		NULL, /* class_data */
		sizeof(ECalBackend),
		0,    /* n_preallocs */
		(GInstanceInitFunc)e_cal_backend_brutus_factory_instance_init
	};

	d("");
	type = g_type_module_register_type(module,
					   E_TYPE_CAL_BACKEND_FACTORY,
					   "ECalBackendBrutusEventsFactory",
					   &info,
					   0);
	return type;
}

static GType
todos_backend_factory_get_type(GTypeModule *module)
{
	GType type;

	GTypeInfo info = {
		sizeof(ECalBackendBrutusFactoryClass),
		NULL, /* base_class_init */
		NULL, /* base_class_finalize */
		(GClassInitFunc)todos_backend_factory_class_init,
		NULL, /* class_finalize */
		NULL, /* class_data */
		sizeof(ECalBackend),
		0,    /* n_preallocs */
		(GInstanceInitFunc)e_cal_backend_brutus_factory_instance_init
	};

	d("");
	type = g_type_module_register_type(module,
					   E_TYPE_CAL_BACKEND_FACTORY,
					   "ECalBackendBrutusTodosFactory",
					   &info,
					   0);
	return type;
}

/* kills the Brutus proxy */
static void
brutus_kill_brutusd(void)
{
	int items = 0;
	pid_t pid;
	gchar *path = g_strconcat(getenv("HOME"), "/", BRUTUSD_PID_FILE_NAME, NULL);
 	FILE *pid_file = fopen(path, "r");
	g_free(path);

	if (!pid_file)
		return;

	items = fscanf(pid_file,"%d", &pid);
	fclose(pid_file);
	if (!items)
		return;

	kill(pid, SIGTERM);
	sleep(1);
	kill(pid, SIGKILL);
}

static GType brutus_types[2];

void
eds_module_initialize(GTypeModule *module)
{
	d("");

	/* create application directory */
	{
		gchar *dir_path = g_strconcat(getenv("HOME"), "/", BRUTUSD_DIR_NAME, NULL);
		int retv = g_mkdir_with_parents(dir_path, 0700);
		g_free(dir_path);

		if ((retv) && (EEXIST != errno)) {
			d("ERROR: Can not get create brutusd directory");
			exit(EXIT_FAILURE);
		}
	}

	/* Start Brutus proxy daemon. The lock file is never closed 
	 * explicitly, only when this process terminates. This is 
	 * _not_ a leak. */
	{
		char *lock_file_path = g_strconcat(getenv("HOME"), "/", BRUTUSD_LOCK_FILE_NAME, NULL);
		int fd = g_open(lock_file_path, O_CREAT | O_WRONLY | O_SYNC, S_IRUSR | S_IWUSR);
		g_free(lock_file_path);
		if (lockf(fd, F_TLOCK, 0)) { // the first to lock will start brutusd 
			d("WARNING: Can not get exclusive lock on brutusd lock file");
		} else {
			d("Exclusive lock on brutusd lock file");

			int retv = system("brutusd");
			dfmt("\"system(brutusd)\" returned %s", (EXIT_SUCCESS == retv) ? "EXIT_SUCCESS" : "EXIT_FAILURE");
			if (EXIT_SUCCESS != retv) {
				d("Could not start brutusd");
				brutus_kill_brutusd();
				return;
			}

			g_atexit(brutus_kill_brutusd);
		}
	}

	/* Start Brutus Keyring daemon */
	{
		int retv = system("brutus-keyringd");
		dfmt("\"system(brutus-keyringd)\" returned %s", (EXIT_SUCCESS == retv) ? "EXIT_SUCCESS" : "EXIT_FAILURE");
		if (EXIT_SUCCESS != retv) {
			d("Could not start brutus-keyringd");
			brutus_kill_brutusd();
			return;
		}
	}
	brutus_types[0] = events_backend_factory_get_type(module);
	brutus_types[1] = todos_backend_factory_get_type(module);
}

void
eds_module_shutdown(void)
{
	d("");
}

void
eds_module_list_types(const GType **types, int *num_types)
{
	d("");
	*types = brutus_types;
	*num_types = 2; // let it be 2 when ORBit2 starts behaving thread wise... 
}
