/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/*
 *    Brutus e-cal backend class.
 *
 *    Authors:
 *        Jules Colding <colding@omesc.com>
 *
 *    Copyright (C) 2006-2007 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *    MA 02111-1307 USA
 */

#ifndef E_CAL_BACKEND_BRUTUS_H
#define E_CAL_BACKEND_BRUTUS_H

#include <libedata-cal/e-cal-backend-sync.h>
#include <libedata-cal/e-cal-backend-cache.h>

#include <idl_output/IMAPISession.h>
#include <idl_output/IMAPIFolder.h>

#include <session/brutus-proxy.h>
#include <session/brutusd-util.h>

G_BEGIN_DECLS

#define E_TYPE_CAL_BACKEND_BRUTUS            (e_cal_backend_brutus_get_type ())
#define E_CAL_BACKEND_BRUTUS(obj)            (G_TYPE_CHECK_INSTANCE_CAST ((obj), E_TYPE_CAL_BACKEND_BRUTUS, ECalBackendBrutus))
#define E_CAL_BACKEND_BRUTUS_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST ((klass), E_TYPE_CAL_BACKEND_BRUTUS, ECalBackendBrutusClass))
#define E_IS_CAL_BACKEND_BRUTUS(obj)         (G_TYPE_CHECK_INSTANCE_TYPE ((obj), E_TYPE_CAL_BACKEND_BRUTUS))
#define E_IS_CAL_BACKEND_BRUTUS_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), E_TYPE_CAL_BACKEND_BRUTUS))

typedef struct _ECalBackendBrutus ECalBackendBrutus;
typedef struct _ECalBackendBrutusClass ECalBackendBrutusClass;

/* Private part of the CalBackendBrutus structure */
typedef struct _ECalBackendBrutusPrivate {
	GMutex *poll_mutex;
	GMutex *connect_mutex;

	gboolean constructed; // initialized with ORB thread and all

	gboolean read_only; // read-only mode if TRUE
	
	CalMode mode;

	CORBA_ORB orb;                    // store-global ORB
	PortableServer_POA root_poa;      // store-global POA

	gboolean use_ssliop; // SSLIOP or not?

	CORBA_char *mapi_profile;
	BRUTUS_IMAPISession mapi_session; // store-global IMAPISession
	CORBA_unsigned_long instance_id;  // IMAPISession instance id

	BRUTUS_IMAPIFolder mapi_folder; // MAPI calendar/tasks folder

	BRUTUS_BrutusProxy brutus_proxy;   // IPv4/6 capable Brutus proxy
	struct BrutusCheckProxy lifeline;  // lifeline callback, local and proxy

	long exchange_time_zone; // The Exchange Server timezone offset from GMT in minutes 
	long client_time_zone;   // The evolution-brutus timezone offset from GMT in minutes 

	long auto_check_timeout; // auto check timeout interval in minutes
	guint timeout_id; // ID of timing source

	gchar *brutus_logon_str; // corbaloc IOR of remote BrutusLogOn object

	ECalBackendCache *cache; // local calender cache

	gchar *email_address; // eamil address of account owner

	gchar *cache_dir; // cache directory

	gchar *brutus_server;

	icaltimezone *default_timezone; 
#if (EDS_EPOCH >= 18)
	ECalSourceType ecal_source_type;
#endif
} ECalBackendBrutusPrivate;

struct _ECalBackendBrutus {
	ECalBackendSync backend;

	/* Private data */
	ECalBackendBrutusPrivate *priv;
};

struct _ECalBackendBrutusClass {
	ECalBackendSyncClass parent_class;
};

extern GType e_cal_backend_brutus_get_type(void);

G_END_DECLS

#endif
