/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/*
 *    Camel Brutus folder summary class.
 *
 *    Authors:
 *        Jules Colding <colding@omesc.com>
 *
 *    Copyright (C) 2005-2007 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *    MA 02111-1307 USA
 */

#ifndef CAMEL_BRUTUS_FOLDER_SUMMARY_H
#define CAMEL_BRUTUS_FOLDER_SUMMARY_H 1

#include <camel/camel-folder.h>
#include <camel/camel-folder-summary.h>

G_BEGIN_DECLS

#define FOLDER_SUMMARIES "folder_summaries"

#define CAMEL_BRUTUS_SUMMARY_TYPE         (camel_brutus_summary_get_type ())
#define CAMEL_BRUTUS_SUMMARY(obj)         CAMEL_CHECK_CAST (obj, camel_brutus_summary_get_type (), CamelBrutusSummary)
#define CAMEL_BRUTUS_SUMMARY_CLASS(klass) CAMEL_CHECK_CLASS_CAST (klass, camel_brutus_summary_get_type (), CamelBrutusSummaryClass)
#define CAMEL_IS_BRUTUS_SUMMARY(obj)      CAMEL_CHECK_TYPE (obj, camel_brutus_summary_get_type ())

typedef struct _CamelBrutusSummary CamelBrutusSummary;
typedef struct _CamelBrutusSummaryClass CamelBrutusSummaryClass;
typedef struct _CamelBrutusMessageInfo CamelBrutusMessageInfo;
typedef struct _CamelBrutusMessageContentInfo CamelBrutusMessageContentInfo ;

struct _CamelBrutusMessageInfo {
	CamelMessageInfoBase info; // base class

	gchar *pr_record_key;      // PR_RECORD_KEY of the remote corresponding message object
};

struct _CamelBrutusMessageContentInfo {
	CamelMessageContentInfo info;
};

// whether message infos in the summary has long-term entryids
#define BRUTUS_FOLDER_SUMMARY_HAS_LONG_TERM_ENTRYIDS (0x00000001)

// sync folder for offline access?
#define BRUTUS_FOLDER_SYNC_OFFLINE                   (0x00000002)

struct _CamelBrutusSummary {
	CamelFolderSummary parent; // base class

	guint32 version;           // structure version tag
	guint32 flags;             // BRUTUS_FOLDER_SUMMARY_ bitmap - present in version 2 or later of the structure

	char session_key[37];      // used to distinguish different run-time sessions - present in version 2 or later of the structure
};

struct _CamelBrutusSummaryClass {
	CamelFolderSummaryClass parent_class;

};

extern void brutus_update_folder_summary_post_light(gpointer data,
						    gpointer static_data);

extern CamelFolderSummary *camel_brutus_summary_new(CamelFolder *folder,
						    const char *summary_filename);

extern void brutus_update_folder_summary(CamelFolder *folder,
					 gboolean force_update, // if TRUE forcefully update everything otherwise only the changes since last the update
					 CamelException *ex);

extern CamelType camel_brutus_summary_get_type(void);


G_END_DECLS

#endif /* CAMEL_BRUTUS_FOLDER_SUMMARY_H */
