/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/*
 *    Authors: Jules Colding <colding@omesc.com>
 *
 *    Camel Brutus folder class.
 *
 *    Copyright (C) 2005-2007 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU General Public License as
 *    published by the Free Software Foundation; either version 2 of
 *    the License, or (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *    MA 02111-1307 USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* system includes */
#include <errno.h>
#include <time.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

/* camel includes */
#include <camel/camel-folder.h>
#include <camel/camel-folder-search.h>
#include <camel/camel-store.h>
#include <camel/camel-i18n.h>
#include <camel/camel-session.h>
#include <camel/camel-exception.h>
#include <camel/camel-stream-mem.h>

#include <libedataserver/e-iconv.h>
#include <libedataserver/md5-utils.h>

/* camel-brutus includes */
#include "camel-brutus-folder.h"
#include "camel-brutus-folder-summary.h"
#include "camel-brutus-store.h"
#include "camel-brutus-private.h"

/* pure brutus includes */
#include <idl_output/IMAPISession.h>
#include <idl_output/return_codes.h>
#include <idl_output/types.h>
#include <idl_output/uid.h>

/* evolution-brutus includes */
#include <server/brutus_conv.h>
#include <server/brutus_mapi.h>
#include <server/brutus_smtp.h>
#include <server/brutus_cpid.h>
#include <server/brutus_util.h>
#include <server/brutus_camel.h>
#include <server/brutus_corba.h>
#include <server/brutus.h>

static CamelOfflineFolderClass *parent_class = NULL;

/*
 * helper functions
 */

/* static CamelMimeMessage* */
/* brutus_foo(void) */
/* { */
/*	CamelMimeMessage *msg; */
/*	CamelInternetAddress *addr; */
/*	const char *text; */
/*	CamelMultipart *mp; */
/*	CamelMimePart *part; */
/*	CamelContentType *ct = NULL; */
/*	char *content_type = NULL; */

/*	ct = camel_content_type_new("text", "plain"); */
/*	camel_content_type_set_param(ct, "charset", "UTF-8"); */
/*	content_type = camel_content_type_format(ct); */

/*	msg = camel_mime_message_new(); */
/*	camel_mime_message_set_subject(msg, "Talking to myself again"); */

/*	addr = camel_internet_address_new(); */
/*	camel_internet_address_add(addr, "NotZed", "notzed@somewhere.com"); */
/*	camel_mime_message_set_from(msg, addr); */
/*	camel_address_remove((CamelAddress*)addr, -1); */

/*	camel_internet_address_add(addr, "NotZed", "notzed@somewhere.else.com"); */
/*	camel_mime_message_set_recipients(msg, "To", addr); */
/*	camel_object_unref(addr); */

/*	mp = camel_multipart_new(); */
/*	camel_data_wrapper_set_mime_type(CAMEL_DATA_WRAPPER(mp), "multipart/mixed"); */
/*	camel_multipart_set_boundary(mp, NULL); */
/*	camel_multipart_set_preface(mp, "This is a message in MIME format.  Get with the decade\n"); */
/*	camel_multipart_set_postface(mp, "\nSome useless after-message text you should never see\n\n"); */

/*	part = camel_mime_part_new(); */
/*	text = "Hello NotZed,\nTalking to yourself again?\n\n  NotZed\n"; */
/*	camel_mime_part_set_content(part, text, strlen(text), "text/plain"); */
/*	camel_mime_part_set_content_type(part, content_type); */
/*	camel_content_type_unref(ct); */
/*	camel_multipart_add_part(mp, part); */
/*	camel_object_unref(part); */

/* /\*	part = camel_mime_part_new(); *\/ */
/* /\*	text = "<html>\n<body>\nHello NotZed,<br>\nTalking to yourself again?<br>\n&nbsp;&nbsp;NotZed<br>\n</body>\n</html>\n"; *\/ */
/* /\*	camel_mime_part_set_content(part, text, strlen(text), "text/html"); *\/ */
/* /\*	camel_multipart_add_part(mp, part); *\/ */
/* /\*	camel_object_unref(part); *\/ */

/*	camel_medium_set_content_object(CAMEL_MEDIUM(msg), CAMEL_DATA_WRAPPER(mp)); */

/*	return msg; */
/* } */



/*
 * Converts an object of type MAPI_MESSAGE to a CamelMimeMessage
 */
static CamelMimeMessage*
brutus_mapi_message_to_camel_mime_message(CamelFolder *folder,
					  const char *uid,
					  CamelException *ex)
{
	CamelMimeMessage *msg = NULL;
	CamelMimePart *mime_part = NULL;
	CamelMultipart *multipart = NULL;
	CamelMessageInfoBase *info = NULL;
	CamelBrutusFolder *brutus_folder = CAMEL_BRUTUS_FOLDER(folder);
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(camel_folder_get_parent_store(folder));
	CamelBrutusStorePrivate *priv = brutus_store->priv;
	CamelInternetAddress *from_addr = NULL;
	CamelInternetAddress *to_addr = NULL;
	CamelInternetAddress *cc_addr = NULL;
	CamelInternetAddress *bcc_addr = NULL;
	CamelContentType *ct = NULL;
	CamelDataWrapper *dw = NULL;
	CamelStream *attach_stream = NULL;
	const char *charset = NULL;
	const char *mime_type = NULL;
	char *content_type = NULL;
	char *long_term_uid = NULL;
	char *msg_id = NULL;
	char *tmp = NULL;
	digest_str_t msg_cache_id = { '\0' };
	gchar *tmp_name = NULL;
	gchar *tmp_value = NULL;
	time_t tmp_time;
	gboolean inbound;
	gboolean plain_text;
	int p = 0;
	CORBA_long n;
	CORBA_long m;
	CORBA_long k;
	CORBA_long g;
	CORBA_long i;
	CORBA_long attach_num;
	CORBA_unsigned_long r;
	CORBA_boolean cb = CORBA_TRUE;
	BRUTUS_IMessage mapi_message = CORBA_OBJECT_NIL;
	BRUTUS_IMAPITable attachment_table = CORBA_OBJECT_NIL;
	BRUTUS_IAttach mapi_attach = CORBA_OBJECT_NIL;
	BRUTUS_seq_octet *attach_data = NULL;
	BRUTUS_BDEFINE obj_type;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_ENTRYID *eid = NULL;
	BRUTUS_ENTRYID *sender_eid = NULL;
	BRUTUS_SPropValue *pv = NULL;
	BRUTUS_seq_SPropValue *prop_values = NULL;
	BRUTUS_seq_SPropValue *msg_prop_values = NULL;
	BRUTUS_seq_SPropValue *attach_prop_values = NULL;
	BRUTUS_seq_SPropValue *ex_msg_prop_values = NULL;
	BRUTUS_SRowSet *attach_row_set = NULL;
	BRUTUS_SRowSet *recp_row_set = NULL;
	CORBA_Environment ev[1];
	BRUTUS_SRowSet *row_set = NULL;                       // for the attachments table
	static BRUTUS_SSortOrderSet sort_order = {            // for the attachments table
		.cCategories = 0,
		.cExpanded = 0,
		.aSort._maximum = 0,
		.aSort._length = 0,
		.aSort._buffer = NULL,
	};
#ifdef BRUTUS_SPY
	char *smtp_header = NULL;
	struct _camel_header_raw *camel_header_raw = NULL;

	static BRUTUS_BDEFINE attach_object_props[8] = {             // for the attachment object
		/* 0 */ BRUTUS_BRUTUS_PR_DISPLAY_NAME,
		/* 1 */ BRUTUS_BRUTUS_PR_ATTACH_FILENAME,
		/* 2 */ BRUTUS_BRUTUS_PR_ATTACH_LONG_FILENAME,
		/* 3 */ BRUTUS_BRUTUS_PR_ATTACH_CONTENT_ID,
		/* 4 */ BRUTUS_BRUTUS_PR_ATTACH_CONTENT_LOCATION,
		/* 5 */ BRUTUS_BRUTUS_PR_ATTACH_MIME_TAG,
		/* 6 */ BRUTUS_BRUTUS_PR_ATTACHMENT_HIDDEN,
		/* 7 */ BRUTUS_BRUTUS_PR_ATTACH_METHOD,
		/* 8 */ /* BRUTUS_BRUTUS_PR_, */
	};
	static BRUTUS_SPropTagArray attach_object_proptag_array = {  // for the attachment object
		._maximum = 8,
		._length = 8,
		._buffer = attach_object_props,
	};
	// get everything
	static BRUTUS_SPropTagArray attach_proptag_array = {         // for the attachment table
		._maximum = 0,
		._length = 0,
		._buffer = NULL,
	};
	static BRUTUS_SPropTagArray recp_proptag_array = {           // for the recipients table
		._maximum = 0,
		._length = 0,
		._buffer = NULL,
	};
	static BRUTUS_SPropTagArray msg_proptag_array = {            // for the message object
		._maximum = 0,
		._length = 0,
		._buffer = NULL,
	};
#else
	CamelStream *cache_stream = NULL;
	CamelStream *msg_stream = NULL;

	// get selected properties
	static BRUTUS_BDEFINE attach_props[1] = {                    // for the attachment table
		/* 0 */ BRUTUS_BRUTUS_PR_ATTACH_NUM,
	};
	static BRUTUS_SPropTagArray attach_proptag_array = {
		._maximum = 1,
		._length = 1,
		._buffer = attach_props,
	};
	static BRUTUS_BDEFINE attach_object_props[8] = {             // for the attachment object
		/* 0 */ BRUTUS_BRUTUS_PR_DISPLAY_NAME,
		/* 1 */ BRUTUS_BRUTUS_PR_ATTACH_FILENAME,
		/* 2 */ BRUTUS_BRUTUS_PR_ATTACH_LONG_FILENAME,
		/* 3 */ BRUTUS_BRUTUS_PR_ATTACH_CONTENT_ID,
		/* 4 */ BRUTUS_BRUTUS_PR_ATTACH_CONTENT_LOCATION,
		/* 5 */ BRUTUS_BRUTUS_PR_ATTACH_MIME_TAG,
		/* 6 */ BRUTUS_BRUTUS_PR_ATTACHMENT_HIDDEN,
		/* 7 */ BRUTUS_BRUTUS_PR_ATTACH_METHOD,
		/* 8 */ /* BRUTUS_BRUTUS_PR_, */
	};
	static BRUTUS_SPropTagArray attach_object_proptag_array = {  // for the attachment object
		._maximum = 8,
		._length = 8,
		._buffer = attach_object_props,
	};
	static BRUTUS_BDEFINE recp_props[5] = {                      // for the recipients table
		/* 0 */ BRUTUS_BRUTUS_PR_RECIPIENT_TYPE,                 // MAPI_TO, MAPI_CC, MAPI_BCC
		/* 1 */ BRUTUS_BRUTUS_PR_DISPLAY_NAME,
		/* 2 */ BRUTUS_BRUTUS_PR_EMAIL_ADDRESS,
		/* 3 */ BRUTUS_BRUTUS_PR_SMTP_ADDRESS,
		/* 4 */ BRUTUS_BRUTUS_PR_ADDRTYPE,                       // (==EX => email=PR_SMTP_ADDRESS) (==SMTP => email=PR_EMAIL_ADDRESS)
		/* 5 */ /* BRUTUS_BRUTUS_PR_, */
	};
	static BRUTUS_SPropTagArray recp_proptag_array = {
		._maximum = 5,
		._length = 5,
		._buffer = recp_props,
	};
	static BRUTUS_BDEFINE msg_props[13] = {                // for the message object
		/* 0 */   BRUTUS_BRUTUS_PR_INTERNET_CPID,             // charset encoding of message body
		/* 1 */   BRUTUS_BRUTUS_PR_RECEIVED_BY_ADDRTYPE,      // to deduce if inbound or outbound
		/* 2 */   BRUTUS_BRUTUS_PR_SENDER_NAME,               // display name of sender
		/* 3 */   BRUTUS_BRUTUS_PR_SENDER_EMAIL_ADDRESS,      // email address of sender if PR_SENDER_ADDRTYPE==SMTP
		/* 4 */   BRUTUS_BRUTUS_PR_INTERNET_MESSAGE_ID,       // RFC 2822 Message-Id
		/* 5 */   BRUTUS_BRUTUS_PR_BODY,
		/* 6 */   BRUTUS_BRUTUS_PR_DISPLAY_TO,
		/* 7 */   BRUTUS_BRUTUS_PR_DISPLAY_CC,
		/* 8 */   BRUTUS_BRUTUS_PR_HASATTACH,
		/* 9 */   BRUTUS_BRUTUS_PR_SENDER_ADDRTYPE,           // EX or SMTP?
		/* 10 */  BRUTUS_BRUTUS_PR_SENDER_ENTRYID,            // open and get PR_SMTP_ADDRESS if PR_SENDER_ADDRTYPE==EX
		/* 11 */  BRUTUS_BRUTUS_PR_SUBJECT,
		/* 12 */  BRUTUS_BRUTUS_PR_MESSAGE_DELIVERY_TIME,
		/* 13 */ /* BRUTUS_BRUTUS_PR_, */
	};
	static BRUTUS_SPropTagArray msg_proptag_array = {
		._maximum = 13,
		._length = 13,
		._buffer = msg_props,
	};
#endif
	static BRUTUS_BDEFINE ex_msg_props[2] = {                // for the message object
		/* 0 */ BRUTUS_BRUTUS_PR_BODY,
		/* 1 */ BRUTUS_BRUTUS_PR_BODY_HTML,
		/* 2 *//*  BRUTUS_BRUTUS_PR_, */
	};
	static BRUTUS_SPropTagArray ex_msg_proptag_array = {
		._maximum = 2,
		._length = 2,
		._buffer = ex_msg_props,
	};

	CORBA_exception_init(ev);

	// create ENTRYID
	eid = brutus_string_to_entryid(uid);
	if (!eid)
		goto no_mem;

	// get long-term entryid
	if (brutus_entryid_is_shortterm(eid)) {
		if (!brutus_folder->online)
			goto offline;

		pv = brutus_open_object_get_one_prop(brutus_folder->mapi_session,
						     CORBA_OBJECT_NIL,
						     eid,
						     BRUTUS_BRUTUS_PR_ENTRYID);
		CORBA_free(eid);
		eid = NULL;
		if (!pv)
			goto error;

		brutus_sbinary_to_entryid(&pv->Value._u.bin, &eid);
		CORBA_free(pv);
		pv = NULL;
		if (!eid)
			goto no_mem;

		long_term_uid = brutus_entryid_to_string(eid);
		uid = long_term_uid;
	}

#ifndef BRUTUS_SPY
	// present in cache?
	brutus_get_digest_str(uid, msg_cache_id);
	cache_stream  = camel_data_cache_get(brutus_folder->msg_cache, "cache", msg_cache_id, NULL);
	d("");
	if (cache_stream) {
		d("");
		// write to msg_stream
		msg_stream = camel_stream_mem_new();
		if (!msg_stream) {
			camel_object_unref(CAMEL_OBJECT(cache_stream));
			cache_stream = NULL;
			camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
					    _("No memory"));
			goto out;
		}
		if (-1 == camel_stream_write_to_stream(cache_stream, msg_stream)) {
			camel_object_unref(CAMEL_OBJECT(cache_stream));
			camel_object_unref(CAMEL_OBJECT(msg_stream));
			cache_stream = NULL;
			msg_stream = NULL;
			d(g_strerror(errno));
			camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
					    _("No memory"));
			goto out;
		}
		camel_stream_reset(msg_stream);

		camel_object_unref(CAMEL_OBJECT(cache_stream));
		cache_stream = NULL;

		// convert into CamelMimeMessage
		msg = camel_mime_message_new();
		if (!msg) {
			camel_object_unref(CAMEL_OBJECT(msg_stream));
			msg_stream = NULL;
			d(g_strerror(errno));
			camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
					    _("No memory"));
			goto out;
		}
		if (-1 == camel_data_wrapper_construct_from_stream((CamelDataWrapper *)msg, msg_stream)) {
			d(g_strerror(errno));
			camel_object_unref(CAMEL_OBJECT(msg));
			msg = NULL;

			if (EINTR == errno) {
				d("");
				camel_exception_setv(ex, CAMEL_EXCEPTION_USER_CANCEL, _("User cancelled"));
			} else {
				d("");
				BRUTUS_FOLDER_LOCK(brutus_folder, cache_mutex);
				camel_data_cache_remove(brutus_folder->msg_cache, "cache", uid, NULL); // might be corrupt, safer to renew later
				BRUTUS_FOLDER_UNLOCK(brutus_folder, cache_mutex);
				camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
						    _("Cannot get the message"));
			}
		}
		camel_object_unref(CAMEL_OBJECT(msg_stream));
		msg_stream = NULL;
		d("");

		if (msg)
			goto out;
		d("");
	}
	d("");
#endif

	// not in cache so we need to be online to get it
	if (!brutus_folder->online)
		goto offline;

	// get selected properties from message object
	br = BRUTUS_IMAPISession_QueryMessage(brutus_folder->mapi_session,
					      eid,
					      &msg_proptag_array,
					      &ex_msg_proptag_array,
					      &recp_proptag_array,
					      &msg_prop_values,
					      &ex_msg_prop_values,
					      &recp_row_set,
					      ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto error;
	}
	if ((BRUTUS_BRUTUS_MAPI_W_ERRORS_RETURNED != br) && (BRUTUS_BRUTUS_S_OK != br)) {
		d(brutus_bresult_to_str(br));
		d("Could not query message object for data");
		goto error;
	}

#ifdef BRUTUS_SPY
	// get full SMTP header if present
	n = brutus_get_spropvalue_index(msg_prop_values, BRUTUS_BRUTUS_PR_TRANSPORT_MESSAGE_HEADERS);
	if (-1 != n) {
		smtp_header = brutus_unfold_rfc822_header(msg_prop_values->_buffer[n].Value._u.lpszA);
		if (smtp_header) {
			camel_header_raw = brutus_get_camel_header_raw(smtp_header);
			free(smtp_header);
			smtp_header = NULL;
		} else {
			d("No memory");
			goto no_mem;
		}
	} else
		camel_header_raw = NULL;
#endif

	// determine origin, inbound or outbound?
	n = brutus_get_spropvalue_index(msg_prop_values, BRUTUS_BRUTUS_PR_RECEIVED_BY_ADDRTYPE); // set by incoming transport provider
	if (-1 != n)
		inbound = TRUE;
	else
		inbound = FALSE; // draft, sent or being edited

	// instantiate mime message
	msg = camel_mime_message_new();
	if (!msg)
		goto no_mem;

	// get info from folder summary
	info = (CamelMessageInfoBase*)camel_folder_summary_uid(folder->summary, uid);
	if (info) {
		d("Message found in summary cache");

		// date
		camel_mime_message_set_date(msg, info->date_sent, brutus_timezone_to_gmt_offset(priv->client_time_zone));

		// subject
		if (info->subject)
			camel_mime_message_set_subject(msg, info->subject);
	} else { // a rare event
		d("Message not in summary cache");

		n = brutus_get_spropvalue_index(msg_prop_values, BRUTUS_BRUTUS_PR_MESSAGE_DELIVERY_TIME);
		if (-1 != n) {
			tmp_time =  brutus_filetime_to_time_t(msg_prop_values->_buffer[n].Value._u.ft);
			camel_mime_message_set_date(msg, tmp_time, brutus_timezone_to_gmt_offset(priv->client_time_zone));
		}

		n = brutus_get_spropvalue_index(msg_prop_values, BRUTUS_BRUTUS_PR_SUBJECT);
		if (-1 != n)
			camel_mime_message_set_subject(msg, msg_prop_values->_buffer[n].Value._u.lpszA);
	}

	/*
	 * Addresses
	 */
	from_addr = camel_internet_address_new();
	if (!from_addr)
		goto no_mem;

	// display name of sender
	n = brutus_get_spropvalue_index(msg_prop_values, BRUTUS_BRUTUS_PR_SENDER_NAME);

	// SMTP address of sender
	i = brutus_get_spropvalue_index(msg_prop_values, BRUTUS_BRUTUS_PR_SENDER_ADDRTYPE);
	if ((-1 != i) && !g_ascii_strcasecmp("SMTP", msg_prop_values->_buffer[i].Value._u.lpszA)) {
		m = brutus_get_spropvalue_index(msg_prop_values, BRUTUS_BRUTUS_PR_SENDER_EMAIL_ADDRESS);
		if (-1 != m)
			tmp = g_strdup(msg_prop_values->_buffer[m].Value._u.lpszA);
		else
			tmp = g_strdup("");
	} else { // open PR_SENDER_ENTRYID and get PR_SMTP_ADDRESS
		i = brutus_get_spropvalue_index(msg_prop_values, BRUTUS_BRUTUS_PR_SENDER_ENTRYID);
		if (-1 != i)
			brutus_sbinary_to_entryid(&msg_prop_values->_buffer[i].Value._u.bin, &sender_eid);
		if (sender_eid) {
			pv = brutus_open_object_get_one_prop(priv->mapi_session,
							     CORBA_OBJECT_NIL,
							     sender_eid,
							     BRUTUS_BRUTUS_PR_SMTP_ADDRESS);
			if ((pv) && (BRUTUS_BRUTUS_PR_SMTP_ADDRESS == pv->ulPropTag)) {
				tmp = g_strdup(pv->Value._u.lpszA);
			} else {
				tmp = g_strdup("");
			}

			CORBA_free(pv);
			pv = NULL;
			CORBA_free(sender_eid);
			sender_eid = NULL;
		} else
			tmp = g_strdup("");
	}
	camel_internet_address_add(from_addr,
				   (-1 != n) ? msg_prop_values->_buffer[n].Value._u.lpszA : "",
				   (const char*)tmp);
	camel_mime_message_set_from(msg, from_addr);
	g_free(tmp);
	tmp = NULL;

	if ((recp_row_set->_length) // anyone?
	    && (BRUTUS_BRUTUS_PT_ERROR != BRUTUS_PROP_TYPE(recp_row_set->_buffer[0]._buffer[0].ulPropTag))) { // we don't know if STORE_SUBMIT_OK is set
		for (k = 0; k<recp_row_set->_length; k++) { // loop all recipients
			n = brutus_get_srow_index(&recp_row_set->_buffer[k], BRUTUS_BRUTUS_PR_RECIPIENT_TYPE);
			m = brutus_get_srow_index(&recp_row_set->_buffer[k], BRUTUS_BRUTUS_PR_DISPLAY_NAME);
			i = brutus_get_srow_index(&recp_row_set->_buffer[k], BRUTUS_BRUTUS_PR_ADDRTYPE); // might be -1

			// (==SMTP => email=PR_EMAIL_ADDRESS) - (==EX => email=PR_SMTP_ADDRESS)
			if (-1 == i) {
				g = -1;
			} else if (!g_ascii_strcasecmp("SMTP", recp_row_set->_buffer[k]._buffer[i].Value._u.lpszA)) {
				g = brutus_get_srow_index(&recp_row_set->_buffer[k], BRUTUS_BRUTUS_PR_EMAIL_ADDRESS);
			} else if (!g_ascii_strcasecmp("EX", recp_row_set->_buffer[k]._buffer[i].Value._u.lpszA)) {
				g = brutus_get_srow_index(&recp_row_set->_buffer[k], BRUTUS_BRUTUS_PR_SMTP_ADDRESS);
			} else
				g = -1;

			if (BRUTUS_BRUTUS_MAPI_TO & recp_row_set->_buffer[k]._buffer[n].Value._u.l) {
				if (!to_addr) {
					to_addr = camel_internet_address_new();
					if (!to_addr)
						goto no_mem;
				}
				camel_internet_address_add(to_addr,
							   (-1 != g) ? recp_row_set->_buffer[k]._buffer[m].Value._u.lpszA : "",  // name  - if PR_[EMAIL/SMTP]_ADDRESS is available use the display name, if not use the empty string
							   (-1 != g) ? recp_row_set->_buffer[k]._buffer[g].Value._u.lpszA : recp_row_set->_buffer[k]._buffer[m].Value._u.lpszA); // email - if PR_[EMAIL/SMTP]_ADDRESS is available use it, otherwise use the PR_DISPLAY_NAME property
				continue;
			}
			if (BRUTUS_BRUTUS_MAPI_CC & recp_row_set->_buffer[k]._buffer[n].Value._u.l) {
				if (!cc_addr) {
					cc_addr = camel_internet_address_new();
					if (!cc_addr)
						goto no_mem;
				}
				camel_internet_address_add(cc_addr,
							   (-1 != g) ? recp_row_set->_buffer[k]._buffer[m].Value._u.lpszA : "",
							   (-1 != g) ? recp_row_set->_buffer[k]._buffer[g].Value._u.lpszA : recp_row_set->_buffer[k]._buffer[m].Value._u.lpszA);
				continue;
			}
			if (BRUTUS_BRUTUS_MAPI_BCC & recp_row_set->_buffer[k]._buffer[n].Value._u.l) {
				if (!bcc_addr) {
					bcc_addr = camel_internet_address_new();
					if (!bcc_addr)
						goto no_mem;
				}
				camel_internet_address_add(bcc_addr,
							   (-1 != g) ? recp_row_set->_buffer[k]._buffer[m].Value._u.lpszA : "",
							   (-1 != g) ? recp_row_set->_buffer[k]._buffer[g].Value._u.lpszA : recp_row_set->_buffer[k]._buffer[m].Value._u.lpszA);
				continue;
			}
		}
	}
	camel_mime_message_set_recipients(msg, CAMEL_RECIPIENT_TYPE_TO, to_addr);
	camel_mime_message_set_recipients(msg, CAMEL_RECIPIENT_TYPE_CC, cc_addr);
	camel_mime_message_set_recipients(msg, CAMEL_RECIPIENT_TYPE_BCC, bcc_addr);

	// Message-Id:
	n = brutus_get_spropvalue_index(msg_prop_values, BRUTUS_BRUTUS_PR_INTERNET_MESSAGE_ID);
	if (-1 != n)
		msg_id = camel_header_msgid_decode(msg_prop_values->_buffer[n].Value._u.lpszA);
	else {
		d("Invalid RFC 822 message header");
		msg_id = NULL;
#ifndef BRUTUS_SPY
		goto error;
#endif
	}
	if (!msg_id) {
		d("No memory or invalid RFC 822 message header");
#ifndef BRUTUS_SPY
		goto error;
#endif
	}
	if (msg_id)
		camel_mime_message_set_message_id(msg, msg_id);


#ifdef BRUTUS_SPY
	// add all MAPI properties as custom X-MAPI- headers
	p = 0;
	do {
		prop_values = p ?  ex_msg_prop_values : msg_prop_values;
		if (2 == p)
			break;
		p++;

		for (n = 0; n < prop_values->_length; n++) {
			if (BRUTUS_BRUTUS_PR_TRANSPORT_MESSAGE_HEADERS == prop_values->_buffer[n].ulPropTag)
				continue;
			tmp_name = g_strjoin(NULL,
					     "X-MAPI-(",
					     brutus_proptype_to_str(BRUTUS_PROP_TYPE(prop_values->_buffer[n].ulPropTag)),
					     ")-(",
					     brutus_proptag_to_str(prop_values->_buffer[n].ulPropTag),
					     ")",
					     NULL);
			if (BRUTUS_BRUTUS_PT_ERROR == BRUTUS_PROP_TYPE(prop_values->_buffer[n].ulPropTag))
				tmp_value = (gchar*)brutus_bresult_to_str((BRUTUS_BRESULT)prop_values->_buffer[n].Value._u.err);
			else
				tmp_value = brutus_print_spropvalue(&prop_values->_buffer[n]);

			camel_medium_set_header(CAMEL_MEDIUM(msg), (const char*)tmp_name, (const char*)tmp_value);
			g_free(tmp_name);
			tmp_name = NULL;

			if (BRUTUS_BRUTUS_PT_ERROR != BRUTUS_PROP_TYPE(prop_values->_buffer[n].ulPropTag))
				g_free(tmp_value);
		}
	} while (TRUE);
	/*
	 * TODO:
	 *
	 * Check if an ePlugin could expand the right-click menu with an additional
	 * property viewing UI, instead of this mess with X- headers.
	 *
	 * Add viewing of RecipientTable and AttachmentTable.
	 */
#else
	p = 0;
	prop_values = NULL;
	tmp_value = NULL;
	tmp_name = NULL;
#endif
	/*
	 * Body part
	 */

	// determine body content
	tmp = NULL;
	plain_text = TRUE;
	if (BRUTUS_BRUTUS_PT_ERROR != BRUTUS_PROP_TYPE(ex_msg_prop_values->_buffer[1].ulPropTag)) { // PR_BODY_HTML
		tmp = g_strdup(ex_msg_prop_values->_buffer[1].Value._u.lpszA);
		plain_text = FALSE;
	} else if (BRUTUS_BRUTUS_PT_ERROR != BRUTUS_PROP_TYPE(ex_msg_prop_values->_buffer[0].ulPropTag)) { // PR_BODY
		tmp = g_strdup(ex_msg_prop_values->_buffer[0].Value._u.lpszA);
	}
	if (!tmp) {
		d("trying IStream to open message body");
		br = BRUTUS_IMAPISession_OpenEntry(brutus_folder->mapi_session,
						   eid,
						   "",
						   BRUTUS_BRUTUS_MAPI_BEST_ACCESS,
						   &obj_type,
						   &mapi_message,
						   ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto error;
		}
		if (BRUTUS_BRUTUS_S_OK != br) {
			d(brutus_bresult_to_str(br));
			d("Could not open message object");
			goto error;
		}

		tmp = brutus_read_text_stream(BRUTUS_BRUTUS_PR_BODY_HTML,
					      mapi_message);
		if (tmp)
			plain_text = FALSE;
		else {
			tmp = brutus_read_text_stream(BRUTUS_BRUTUS_PR_BODY,
						      mapi_message);
			if (!tmp)
				tmp = g_strdup(" "); // must not be the empty - otherwise *BOOM* when loading a message without a body part
		}
	}

	// determine charset encoding on body part
	n = brutus_get_spropvalue_index(msg_prop_values, BRUTUS_BRUTUS_PR_INTERNET_CPID);
	if ((-1 != n) && (BRUTUS_BRUTUS_PT_ERROR != BRUTUS_PROP_TYPE(msg_prop_values->_buffer[n].ulPropTag)))
		charset = brutus_cpid_to_label(msg_prop_values->_buffer[n].Value._u.l);
	charset = charset ? e_iconv_charset_name(charset) : "us-ascii";
	ct = camel_content_type_new("text", plain_text ? "plain" : "html");
	camel_content_type_set_param(ct, "charset", charset);
	charset = NULL;
	content_type = camel_content_type_format(ct);

	n = brutus_get_spropvalue_index(msg_prop_values, BRUTUS_BRUTUS_PR_HASATTACH);
	if ((-1 == n) || !msg_prop_values->_buffer[n].Value._u.b) { // no attachments
		camel_mime_part_set_content(CAMEL_MIME_PART(msg), tmp, strlen(tmp),  plain_text ? "text/plain" : "text/html");
		camel_mime_part_set_content_type(CAMEL_MIME_PART(msg), content_type);
		g_free(tmp);
		tmp = NULL;
	} else { // one or more attachments present
		multipart = camel_multipart_new();
		if (!multipart)
			goto no_mem;
		camel_data_wrapper_set_mime_type(CAMEL_DATA_WRAPPER(multipart), "multipart/mixed"); // or alternative of both PR_BODY and PR_BODY_HTML exists?
		camel_multipart_set_boundary(multipart, NULL);

		// create a mime part for the body text
		mime_part = camel_mime_part_new();
		if (!mime_part)
			goto no_mem;
		camel_mime_part_set_encoding(mime_part, CAMEL_TRANSFER_ENCODING_8BIT);

		// set body content
		camel_mime_part_set_content(mime_part, tmp, strlen(tmp), plain_text ? "text/plain" : "text/html");
		camel_mime_part_set_content_type(mime_part, content_type);
		g_free(tmp);
		tmp = NULL;

		// add content part to mime multipart
		camel_multipart_add_part(multipart, mime_part);
		camel_object_unref(CAMEL_OBJECT(mime_part));
		mime_part = NULL;

		/*
		 * Attachments
		 */
		camel_operation_start(NULL, _("Fetching message attachments"));

		// check for already opened message
		cb = CORBA_Object_is_nil(mapi_message, ev);
		if (ORBIT2_EX(ev) || cb) {
			br = BRUTUS_IMAPISession_OpenEntry(brutus_folder->mapi_session,
							   eid,
							   "",
							   BRUTUS_BRUTUS_MAPI_BEST_ACCESS,
							   &obj_type,
							   &mapi_message,
							   ev);
			if (ORBIT2_EX(ev)) {
				PRINT_ORBIT2_EX(ev);
				camel_operation_end(NULL);
				goto error;
			}
			if (BRUTUS_BRUTUS_S_OK != br) {
				d(brutus_bresult_to_str(br));
				d("Could not open message object");
				camel_operation_end(NULL);
				goto error;
			}
		}
		br = BRUTUS_IMessage_GetAttachmentTable(mapi_message,
							0,
							&attachment_table,
							ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			camel_operation_end(NULL);
			goto error;
		}
		if (BRUTUS_BRUTUS_S_OK != br) {
			d(brutus_bresult_to_str(br));
			d("Could not query message object for attachment table");
			camel_operation_end(NULL);
			goto error;
		}

		br = brutus_QueryAllRows(attachment_table,
					 &attach_proptag_array,
					 CORBA_OBJECT_NIL,
					 &sort_order,
					 0,
					 &row_set,
					 ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			camel_operation_end(NULL);
			goto error;
		}
		if (BRUTUS_BRUTUS_S_OK != br) {
			d(brutus_bresult_to_str(br));
			camel_operation_end(NULL);
			goto error;
		}

		// iterate over content (columns)
		// See http://windowssdk.msdn.microsoft.com/library/default.asp?url=/library/en-us/MAPI/html/_mapi1book_opening_an_attachment.asp
		for (r = 0; r < row_set->_length; r++) {
			// progress bar
			camel_operation_progress(NULL, (100*r)/row_set->_length);

			n = brutus_get_srow_index(&row_set->_buffer[r], BRUTUS_BRUTUS_PR_ATTACH_NUM);
			if (-1 == n) {
				d("No PR_ATTACH_NUM");
				continue;
			}
			attach_num = row_set->_buffer[r]._buffer[n].Value._u.l;

			br = BRUTUS_IMessage_OpenAttach(mapi_message,
							attach_num,
							"",
							0, //  read-only
							&mapi_attach,
							ev);
			if (ORBIT2_EX(ev)) {
				PRINT_ORBIT2_EX(ev);
				continue;
			}
			if (BRUTUS_BRUTUS_S_OK != br) {
				d(brutus_bresult_to_str(br));
				continue;
			}

			// get attachment properties
			br = BRUTUS_IMAPIProp_GetProps(mapi_attach,
						       &attach_object_proptag_array,
						       0,
						       &attach_prop_values,
						       ev);
			if (ORBIT2_EX(ev)) {
				PRINT_ORBIT2_EX(ev);
				goto next;
			}
			if ((BRUTUS_BRUTUS_MAPI_W_ERRORS_RETURNED != br) && (BRUTUS_BRUTUS_S_OK != br)) {
				d(brutus_bresult_to_str(br));
				goto next;
			}

			// print attachment method
			n = brutus_get_spropvalue_index(attach_prop_values, BRUTUS_BRUTUS_PR_ATTACH_METHOD);
			if (-1 != n) {
				if (BRUTUS_BRUTUS_PT_ERROR != BRUTUS_PROP_TYPE(attach_prop_values->_buffer[n].ulPropTag)) { // PR_ATTACH_METHOD
					switch (attach_prop_values->_buffer[n].Value._u.l) {
					case BRUTUS_BRUTUS_NO_ATTACHMENT :
						d("PR_ATTACH_METHOD: NO_ATTACHMENT");
						break;
					case BRUTUS_BRUTUS_ATTACH_BY_VALUE :
						d("PR_ATTACH_METHOD: ATTACH_BY_VALUE");
						break;
					case BRUTUS_BRUTUS_ATTACH_BY_REFERENCE :
						d("PR_ATTACH_METHOD: ATTACH_BY_REFERENCE");
						break;
					case BRUTUS_BRUTUS_ATTACH_BY_REF_RESOLVE :
						d("PR_ATTACH_METHOD: ATTACH_BY_REF_RESOLVE");
						break;
					case BRUTUS_BRUTUS_ATTACH_BY_REF_ONLY :
						d("PR_ATTACH_METHOD: ATTACH_BY_REF_ONLY");
						break;
					case BRUTUS_BRUTUS_ATTACH_EMBEDDED_MSG :
						d("PR_ATTACH_METHOD: ATTACH_EMBEDDED_MS");
						break;
					case BRUTUS_BRUTUS_ATTACH_OLE :
						d("PR_ATTACH_METHOD: ATTACH_OLE");
						break;
					default :
						d("PR_ATTACH_METHOD: Unknown");
						goto next;
					}
				}
			}

			// read attachment data from stream
			attach_data = brutus_read_binary_stream(BRUTUS_BRUTUS_PR_ATTACH_DATA_BIN, mapi_attach);
			if (!attach_data)
				goto next;

			// create a mime part for the attachment data
			mime_part = camel_mime_part_new();
			if (!mime_part) {
				d("No memory");
				goto next;
			}

			//
			// set attachment properties
			//

			camel_mime_part_set_encoding(mime_part, CAMEL_TRANSFER_ENCODING_BASE64);

			n = brutus_get_spropvalue_index(attach_prop_values, BRUTUS_BRUTUS_PR_DISPLAY_NAME);
			if (-1 != n) {
				camel_mime_part_set_description(mime_part, attach_prop_values->_buffer[n].Value._u.lpszA);
				d(attach_prop_values->_buffer[n].Value._u.lpszA);
			}

			n = brutus_get_spropvalue_index(attach_prop_values, BRUTUS_BRUTUS_PR_ATTACH_LONG_FILENAME);
			if (-1 != n) {
				camel_mime_part_set_filename(mime_part, attach_prop_values->_buffer[n].Value._u.lpszA);
				d(attach_prop_values->_buffer[n].Value._u.lpszA);
			} else {
				n = brutus_get_spropvalue_index(attach_prop_values, BRUTUS_BRUTUS_PR_ATTACH_FILENAME);
				if (-1 != n) {
					camel_mime_part_set_filename(mime_part, attach_prop_values->_buffer[n].Value._u.lpszA);
					d(attach_prop_values->_buffer[n].Value._u.lpszA);
				}
			}

			n = brutus_get_spropvalue_index(attach_prop_values, BRUTUS_BRUTUS_PR_ATTACH_CONTENT_ID);
			if (-1 != n) {
				camel_mime_part_set_content_id(mime_part, attach_prop_values->_buffer[n].Value._u.lpszA);
				d(attach_prop_values->_buffer[n].Value._u.lpszA);
			}

			n = brutus_get_spropvalue_index(attach_prop_values, BRUTUS_BRUTUS_PR_ATTACH_CONTENT_LOCATION);
			if (-1 != n) {
				camel_mime_part_set_content_location(mime_part, attach_prop_values->_buffer[n].Value._u.lpszA);
				d(attach_prop_values->_buffer[n].Value._u.lpszA);
			}

			n = brutus_get_spropvalue_index(attach_prop_values, BRUTUS_BRUTUS_PR_ATTACH_MIME_TAG);
			if (-1 != n) {
				d(attach_prop_values->_buffer[n].Value._u.lpszA);
				mime_type = attach_prop_values->_buffer[n].Value._u.lpszA;
			} else {
				mime_type = "application/octet-stream"; // hopefully safe default
			}

			n = brutus_get_spropvalue_index(attach_prop_values, BRUTUS_BRUTUS_PR_ATTACHMENT_HIDDEN);
			if (-1 != n) {
				if (attach_prop_values->_buffer[n].Value._u.b)
					camel_mime_part_set_disposition(mime_part, "inline");
				d(attach_prop_values->_buffer[n].Value._u.b ? "TRUE" : "FALSE");
			}

			//
			// set attachment data
			//
			dw = camel_data_wrapper_new();
			if (!dw) {
				d("No memory");
				goto next;
			}
			camel_data_wrapper_set_mime_type(dw, mime_type);
			attach_stream = camel_stream_mem_new_with_buffer((const char*)attach_data->_buffer, attach_data->_length);
			camel_stream_reset(attach_stream);
			camel_data_wrapper_construct_from_stream(dw, attach_stream);
			camel_object_unref(CAMEL_OBJECT(attach_stream));
			attach_stream = NULL;

			camel_medium_set_content_object(CAMEL_MEDIUM(mime_part), dw);
			camel_object_unref(CAMEL_OBJECT(dw));
			dw = NULL;

			// add content part to mime multipart
			camel_multipart_add_part(multipart, mime_part);
			camel_object_unref(CAMEL_OBJECT(mime_part));
			mime_part = NULL;

		next:
			if (attach_stream) {
				camel_object_unref(CAMEL_OBJECT(attach_stream));
				attach_stream = NULL;
			}

			if (dw) {
				camel_object_unref(CAMEL_OBJECT(dw));
				dw = NULL;
			}

			brutus_release_object(BRUTUS_IAttach_tc, &mapi_attach, CORBA_OBJECT_NIL, ev);
			PRINT_IF_ORBIT2_EX(ev);

			CORBA_free(attach_prop_values);
			attach_prop_values = NULL;

			CORBA_free(attach_data);
			attach_data = NULL;
		}
		camel_operation_progress(NULL, 100);

		camel_medium_set_content_object(CAMEL_MEDIUM(msg), CAMEL_DATA_WRAPPER(multipart));
		camel_operation_end(NULL);
	}
	camel_content_type_unref(ct);
	g_free(content_type);


#ifndef BRUTUS_SPY
	// add to cache
	d(uid);
	BRUTUS_FOLDER_LOCK(brutus_folder, cache_mutex);
	cache_stream = camel_data_cache_add(brutus_folder->msg_cache, "cache", msg_cache_id, NULL);
	d(g_strerror(errno));
	if (cache_stream) {
		d("");
		if (-1 == camel_data_wrapper_write_to_stream((CamelDataWrapper *)msg, cache_stream)) {
			d(g_strerror(errno));
			camel_data_cache_remove(brutus_folder->msg_cache, "cache", msg_cache_id, NULL); // no space on device?
		} else if (-1 == camel_stream_flush(cache_stream)) {
			d(g_strerror(errno));
			camel_data_cache_remove(brutus_folder->msg_cache, "cache", msg_cache_id, NULL); // no space on device?
		}
		dfmt("message written to message cache: %sd", uid);
		camel_stream_close(cache_stream);
		camel_object_unref(CAMEL_OBJECT(cache_stream));
		cache_stream = NULL;
	}
	d("");
	BRUTUS_FOLDER_UNLOCK(brutus_folder, cache_mutex);
#endif
	goto out;

offline:
	camel_exception_set(ex, CAMEL_EXCEPTION_SERVICE_NOT_CONNECTED,
			    _("Could not retrieve message - The folder is offline"));
	goto out;

error:
	camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
			    _("Could not construct the message - Error"));
	goto out;

no_mem:
	camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
			    _("Could not construct the message - No memory"));
	goto out;

out:
	brutus_release_object(BRUTUS_IMAPITable_tc, &attachment_table, CORBA_OBJECT_NIL, ev);
	PRINT_IF_ORBIT2_EX(ev);

	brutus_release_object(BRUTUS_IMessage_tc, &mapi_message, CORBA_OBJECT_NIL, ev);
	PRINT_IF_ORBIT2_EX(ev);

	if (mime_part)
		camel_object_unref(CAMEL_OBJECT(mime_part));

	if (multipart)
		camel_object_unref(CAMEL_OBJECT(multipart));

	if (camel_exception_is_set(ex) && msg) {
		camel_object_unref(CAMEL_OBJECT(msg));
		msg = NULL;
	}

	if (from_addr)
		camel_object_unref(CAMEL_OBJECT(from_addr));

	if (to_addr)
		camel_object_unref(CAMEL_OBJECT(to_addr));

	if (cc_addr)
		camel_object_unref(CAMEL_OBJECT(cc_addr));

	if (bcc_addr)
		camel_object_unref(CAMEL_OBJECT(bcc_addr));

	if (info)
		camel_message_info_free(info);

	free(msg_id);
	free(long_term_uid);

	CORBA_free(eid);
	CORBA_free(msg_prop_values);
	CORBA_free(ex_msg_prop_values);
	CORBA_free(attach_row_set);
	CORBA_free(recp_row_set);
	CORBA_exception_free(ev);

#ifdef BRUTUS_SPY
	if (camel_header_raw)
		camel_header_raw_clear(&camel_header_raw);
#endif

#ifdef BRUTUS_DEBUG
	if (msg)
		camel_mime_message_dump(msg, 1); // this leaks memory but we don't care as we are in debug mode...
#endif
	return msg;
}


/*
 * virtual methods
 */

static int brutus_folder_getv(CamelObject *object,
			      CamelException *ex,
			      CamelArgGetV *args);

void camel_brutus_folder_sync(CamelFolder *folder,
			      gboolean expunge,
			      CamelException *ex);

void camel_brutus_folder_expunge(CamelFolder *folder,
				 CamelException *ex);
/* message manipulation */
void camel_brutus_folder_append_message(CamelFolder *folder,
					CamelMimeMessage *message,
					const CamelMessageInfo *info,
					char **appended_uid,
					CamelException *ex);
/* uid based access operations */
CamelMimeMessage *camel_brutus_folder_get_message(CamelFolder *folder,
						  const char *uid,
						  CamelException *ex);
/* search api */
gboolean camel_brutus_folder_has_search_capability(CamelFolder *folder);

/* GPtrArray *camel_brutus_folder_search_by_expression(CamelFolder *folder, */
/*						    const char *expr, */
/*						    CamelException *ex); */

/* GPtrArray *camel_brutus_folder_search_by_uids(CamelFolder *folder, */
/*					      const char *expr, */
/*					      GPtrArray *uids, */
/*					      CamelException *ex); */

/* void camel_brutus_folder_search_free(CamelFolder *folder, */
/*				     GPtrArray *result); */

/* summary info */
void camel_brutus_folder_refresh_info(CamelFolder *folder,
				      CamelException *ex);

void camel_brutus_folder_transfer_messages_to(CamelFolder *source,
					      GPtrArray *uids,
					      CamelFolder *dest,
					      GPtrArray **transferred_uids,
					      gboolean delete_originals,
					      CamelException *ex);

/******************************
 *    folder caching thread   *
 *****************************/

static struct timespec offline_check_delay = {
	.tv_sec = 1,
	.tv_nsec = 0,
};

/*
 * This thread check for the sync_offline checkbox on the folder
 * properties dialog to be checked. It will cache messages in the
 * folder if the checkbox is checked whenever the folder summary
 * are ready to be queried. No action is taken if the checkbox
 * is un-checked. The thread will terminate when one cache-run has
 * been completed.
 */
static gpointer
cache_folder_offline_thread_func(gpointer data)
{
	gboolean do_break = FALSE;
	gboolean cop_unreg = FALSE;
	CamelFolder *folder =  CAMEL_FOLDER(data);
	CamelBrutusFolder *brutus_folder = CAMEL_BRUTUS_FOLDER(folder);
	CamelBrutusSummary *cbs = CAMEL_BRUTUS_SUMMARY(folder->summary);
	GPtrArray *summary = NULL;
	CamelBrutusMessageInfo *msg_info = NULL;
	CamelMimeMessage *msg;
	CamelException ex;
	CamelOperation *cop = NULL;
	CORBA_unsigned_long r = 0;

	// offline?
	if (!brutus_is_connected(brutus_folder->brutus_server,
				 brutus_folder->mapi_session))
		return NULL;

	cop = camel_operation_registered();
	if (!cop) {
		cop = camel_operation_new(NULL, NULL);
		if (!cop) {
			d("");
			return NULL;
		}
		camel_operation_register(cop);
		cop_unreg = TRUE;
	}

	while (TRUE) {
		nanosleep(&offline_check_delay, NULL);

		BRUTUS_FOLDER_TRY_LOCK(brutus_folder, finalize_mutex);
		if (!(((CamelOfflineFolder*)folder)->sync_offline))
			goto unlock;

		BRUTUS_FOLDER_LOCK(brutus_folder, summary_mutex);
		cbs = CAMEL_BRUTUS_SUMMARY(folder->summary);
		if (!(cbs->flags & BRUTUS_FOLDER_SUMMARY_HAS_LONG_TERM_ENTRYIDS)) { // does the summary contain long-term entryids?
			if (strcmp(cbs->session_key, brutus_folder->session_key)) { // the short-term entryids are *not* valid
				BRUTUS_FOLDER_UNLOCK(brutus_folder, summary_mutex);
				goto unlock;
			}
		}
		BRUTUS_FOLDER_UNLOCK(brutus_folder, summary_mutex);

		summary = camel_folder_get_summary(folder);
		if (!summary) {
			do_break = TRUE;
			goto unlock;
		}

		camel_operation_start(cop, _("Caching remote messages locally from %"), folder->name);
		camel_operation_progress(cop, 0);
		for (r = 0; r < summary->len; r++) {

			if (camel_operation_cancel_check(cop)) {
				d("Caching cancelled");
				goto canceled;
			}

			if (!brutus_is_connected(brutus_folder->brutus_server,
						 brutus_folder->mapi_session))
				break;

			// get the cached message info pointer to retrieve the uid
			msg_info = g_ptr_array_index(summary, r);

			// this will update the cached copy of the cached message
			msg = camel_brutus_folder_get_message(folder, msg_info->info.uid, &ex);
			if (msg)
				camel_object_unref(CAMEL_OBJECT(msg));

			camel_operation_progress(cop, (100*(r+1))/summary->len);
		}
		if (summary)
			camel_folder_free_summary(folder, summary);

	canceled:
		camel_operation_progress(cop, 100);
		camel_operation_end(cop);

		do_break = TRUE;
	unlock:
		BRUTUS_FOLDER_UNLOCK(brutus_folder, finalize_mutex);
		if (do_break)
			break;
	}

	if (cop_unreg)
		camel_operation_unregister(cop);
	camel_operation_unref(cop);
out:
	return NULL;
}


/******************************
 *       Implementations      *
 *****************************/

void
camel_brutus_folder_refresh_info(CamelFolder *folder,
				 CamelException *ex)
{
	CamelBrutusFolder *brutus_folder = CAMEL_BRUTUS_FOLDER(folder);

	d("");
	if (!brutus_folder->online) {
		camel_exception_set(ex, CAMEL_EXCEPTION_SERVICE_NOT_CONNECTED,
				    _("Not connected - The folder is read-only"));
		return;
	}

	// incremental update only is the default case
	brutus_update_folder_summary(folder, FALSE, ex);
}

void camel_brutus_folder_expunge(CamelFolder *folder,
				 CamelException *ex)
{
	CamelBrutusFolder *brutus_folder = CAMEL_BRUTUS_FOLDER(folder);

	d("");
	if (!brutus_folder->online) {
		camel_exception_set(ex, CAMEL_EXCEPTION_SERVICE_NOT_CONNECTED,
				    _("Not connected - The folder is read-only"));
		return;
	}

	camel_brutus_folder_sync(folder, TRUE, ex);
}

void camel_brutus_folder_sync(CamelFolder *folder,
			      gboolean expunge,
			      CamelException *ex)
{
	int n;
	CamelBrutusFolder *brutus_folder = CAMEL_BRUTUS_FOLDER(folder);
	CamelBrutusMessageInfo *bmi = NULL;
	CamelFolderChangeInfo *changes = NULL;
	GPtrArray *expunge_msg_infos = NULL;
	gboolean summary_dirty = FALSE;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_BDEFINE msg_flags = 0;
	BRUTUS_BDEFINE old_msg_flags;
	const BRUTUS_BDEFINE mask = BRUTUS_BRUTUS_MSGSTATUS_DELMARKED | BRUTUS_BRUTUS_MSGSTATUS_HIGHLIGHTED;
	CORBA_Environment ev[1];
	BRUTUS_ENTRYID *eid = NULL;
	BRUTUS_seq_ENTRYID msg_id_list = {
		._maximum = 0,
		._length = 0,
		._buffer = NULL,
	};

	d(folder->full_name);

	if (!brutus_folder->is_a_folder)
		return;

	if (!brutus_folder->online) {
		camel_exception_set(ex, CAMEL_EXCEPTION_SERVICE_NOT_CONNECTED,
				    _("Not connected - The folder is read-only"));
		return;
	}

	CORBA_exception_init(ev);

	changes = camel_folder_change_info_new();
	if (!changes)
		goto no_mem;

	expunge_msg_infos = g_ptr_array_new();
	if (!expunge_msg_infos)
		goto no_mem;

	camel_operation_start(NULL, _("Synchronizing folder"));

	for (n = 0; n < folder->summary->messages->len; n++) {
		msg_flags = 0;
		bmi = (CamelBrutusMessageInfo*)g_ptr_array_index(folder->summary->messages, n);

		// there might bo non-deleted message left since last logon that require an action now
		if (bmi->info.flags & CAMEL_MESSAGE_DELETED) {
			msg_flags |= BRUTUS_BRUTUS_MSGSTATUS_DELMARKED;
			if (expunge)
				g_ptr_array_add(expunge_msg_infos, (gpointer)camel_message_info_uid(&bmi->info));
		}

		// we can safely skip further processing now if the info hasn't changed
		if (!(bmi->info.flags & CAMEL_MESSAGE_FOLDER_FLAGGED))
			continue;
		else
			summary_dirty = TRUE;

		if (bmi->info.flags & CAMEL_MESSAGE_FLAGGED)
			msg_flags |= BRUTUS_BRUTUS_MSGSTATUS_HIGHLIGHTED;

		// now set the flags on the remote message
		eid = brutus_string_to_entryid(camel_message_info_uid(&bmi->info));
		if (!eid) {
			d("No memory or corrupted uid");
			continue;
		}
		br = BRUTUS_IMAPIFolder_SetMessageStatus(brutus_folder->mapi_folder,
							 eid,
							 msg_flags,
							 mask,
							 &old_msg_flags,
							 ev);
		CORBA_free(eid);
		eid = NULL;
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			camel_exception_set(ex, CAMEL_EXCEPTION_SERVICE_UNAVAILABLE,
					    _("Could not set new message status: Communication error"));
			goto quit;
		}
		if (BRUTUS_BRUTUS_S_OK != br) {
			d(brutus_bresult_to_str(br));
			camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
					    _("Could not set new message status"));
			continue;
		}

		// set or clear the read flag
		if (!brutus_string_to_seq_entryid(camel_message_info_uid(&bmi->info), &msg_id_list)) {
			d("No memory or corrupted uid");
			continue;
		}
		if (bmi->info.flags & CAMEL_MESSAGE_SEEN) {
			d("");
			br = BRUTUS_IMAPIFolder_SetReadFlags(brutus_folder->mapi_folder, &msg_id_list, CORBA_OBJECT_NIL, 0, ev);
		} else {
			d("");
			br = BRUTUS_IMAPIFolder_SetReadFlags(brutus_folder->mapi_folder, &msg_id_list, CORBA_OBJECT_NIL, BRUTUS_BRUTUS_CLEAR_READ_FLAG, ev);
		}
		d(brutus_bresult_to_str(br));
		brutus_free_seq_entryid_content(&msg_id_list);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			camel_exception_set(ex, CAMEL_EXCEPTION_SERVICE_UNAVAILABLE,
					    _("Could not set new message status: Communication error"));
			goto quit;
		}
		if (BRUTUS_BRUTUS_S_OK != br) {
			d(brutus_bresult_to_str(br));
			d("Could not set the READ flag");
			continue;
		} else
			d("");

		// all is well - unflag this message info
		bmi->info.flags &= ~CAMEL_MESSAGE_FOLDER_FLAGGED;
	}

	camel_operation_end(NULL);

	if (expunge) {
		camel_operation_start(NULL, _("Expunging folder"));

		if (!expunge_msg_infos->len)
			goto quit;

		if (!brutus_uid_array_to_seq_entryid(expunge_msg_infos, &msg_id_list)) {
			camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
					    _("Message IDs are garbled"));
			goto quit;
		}

		br = BRUTUS_IMAPIFolder_DeleteMessages(brutus_folder->mapi_folder,
						       &msg_id_list,
						       CORBA_OBJECT_NIL,
						       0,
						       ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			camel_exception_set(ex, CAMEL_EXCEPTION_SERVICE_UNAVAILABLE,
					    _("Could not delete messages: Communication error"));
			goto quit;
		}
		if ((BRUTUS_BRUTUS_MAPI_W_PARTIAL_COMPLETION == br) || (BRUTUS_BRUTUS_S_OK == br)) {
			for (n = 0; n < expunge_msg_infos->len; n++) {
				camel_folder_change_info_remove_uid(changes, (const char*)g_ptr_array_index(expunge_msg_infos, n));
				camel_folder_summary_remove_uid(folder->summary, (const char*)g_ptr_array_index(expunge_msg_infos, n));
			}
		}
		if (BRUTUS_BRUTUS_MAPI_W_PARTIAL_COMPLETION == br) {
			camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
					    _("Some messages could not be deleted"));
			goto quit;
		}
		if (BRUTUS_BRUTUS_S_OK != br) {
			d(brutus_bresult_to_str(br));
			camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
					    _("Could not delete messages"));
			goto quit;
		}
	quit:
		camel_operation_end(NULL);
	}
	goto out;

no_mem:
	camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
			    _("No memory"));
out:
	CORBA_exception_free(ev);

	brutus_free_seq_entryid_content(&msg_id_list);

	if (expunge_msg_infos)
		g_ptr_array_free(expunge_msg_infos, TRUE);

	if (summary_dirty || (changes && camel_folder_change_info_changed(changes))) {
		folder->summary->time = brutus_get_server_utc_time(brutus_folder->brutus_server,
								   brutus_folder->mapi_session, 
								   ((CamelService*)camel_folder_get_parent_store(folder))->url->host);
		camel_folder_summary_touch(folder->summary);
		camel_folder_summary_save(folder->summary);

		camel_object_trigger_event(folder, "folder_changed", changes);
	}
	if (changes)
		camel_folder_change_info_free(changes);

	if (BRUTUS_BRUTUS_MAPI_W_PARTIAL_COMPLETION == br)
		brutus_update_folder_summary(folder, FALSE, ex);
}

/* message manipulation */
void
camel_brutus_folder_append_message(CamelFolder *folder,
				   CamelMimeMessage *message,
				   const CamelMessageInfo *info,
				   char **appended_uid,
				   CamelException *ex)
{
	CamelBrutusFolder *brutus_folder = CAMEL_BRUTUS_FOLDER(folder);
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(camel_folder_get_parent_store(folder));
	CamelBrutusStorePrivate *priv = brutus_store->priv;
	BRUTUS_ENTRYID *eid = NULL;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_IMessage mapi_message = CORBA_OBJECT_NIL;
	CORBA_Environment ev[1];

	d(folder->full_name);

	if (!brutus_folder->is_a_folder)
		return;

	if (!brutus_folder->online) {
		camel_exception_set(ex, CAMEL_EXCEPTION_SERVICE_NOT_CONNECTED,
				    _("Not connected - The folder is read-only"));
		return;
	}

	CORBA_exception_init(ev);

	if (appended_uid)
		*appended_uid = NULL;

	br = BRUTUS_IMAPIFolder_CreateMessage(brutus_folder->mapi_folder, "", 0, &mapi_message, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		camel_exception_set(ex, CAMEL_EXCEPTION_SERVICE_UNAVAILABLE,
				    _("Cannot create message: Communication error"));
		goto exit;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("Cannot create message: Server error"));
		goto out;
	}

	// convert to MAPI
	if (!brutus_camel_mime_message_to_mapi(message, mapi_message, priv->mapi_session, priv->root_poa, NULL, ex))
		goto out;

	br = BRUTUS_IMessage_SaveChanges(mapi_message, BRUTUS_BRUTUS_KEEP_OPEN_READONLY, ev);
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		camel_exception_set(ex, CAMEL_EXCEPTION_SERVICE_UNAVAILABLE,
				    _("Could not submit outgoing message"));
		goto out;
	}

	// get long-term entry id
	if (appended_uid) {
		eid = brutus_get_entryid(mapi_message, BRUTUS_BRUTUS_PR_ENTRYID);
		*appended_uid = brutus_entryid_to_string(eid);
		CORBA_free(eid);
	}

out:
	brutus_release_object(BRUTUS_IMessage_tc, &mapi_message, CORBA_OBJECT_NIL, ev);
	PRINT_IF_ORBIT2_EX(ev);

	brutus_update_folder_summary(folder, FALSE, ex); // slow, synchronous and complete

exit:
	CORBA_exception_free(ev);

	return;
}

/* uid based access operations */
CamelMimeMessage*
camel_brutus_folder_get_message(CamelFolder *folder,
				const char *uid,
				CamelException *ex)
{
	d(folder->full_name);
	return brutus_mapi_message_to_camel_mime_message(folder, uid, ex);
}

/* search api */
gboolean
camel_brutus_folder_has_search_capability(CamelFolder *folder)
{
	d(folder->full_name);

	return FALSE;
}

/*
 * Commented out for now
 */

/* GPtrArray* */
/* camel_brutus_folder_search_by_expression(CamelFolder *folder, */
/*					 const char *expr, */
/*					 CamelException *ex) */
/* { */
/*	GPtrArray *retv = NULL; */

/*	d(""); */

/*	return retv; */
/* } */

/* GPtrArray* */
/* camel_brutus_folder_search_by_uids(CamelFolder *folder, */
/*				   const char *expr, */
/*				   GPtrArray *uids, */
/*				   CamelException *ex) */
/* { */
/*	GPtrArray *retv = NULL; */

/*	d(""); */

/*	return retv; */
/* } */

/* void */
/* camel_brutus_folder_search_free(CamelFolder *folder, */
/*				GPtrArray *result) */
/* { */
/* } */

void
camel_brutus_folder_transfer_messages_to(CamelFolder *source,
					 GPtrArray *uids,
					 CamelFolder *dest,
					 GPtrArray **transferred_uids,
					 gboolean delete_originals,
					 CamelException *ex)
{
	CamelBrutusFolder *source_folder = CAMEL_BRUTUS_FOLDER(source); // using dest ==> noop() - MAPI bug I guess...
	CamelBrutusFolder *dest_folder = CAMEL_BRUTUS_FOLDER(dest); // using dest ==> noop() - MAPI bug I guess...
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(camel_folder_get_parent_store(source));
	CamelBrutusStorePrivate *priv = brutus_store->priv;
	const char *mv_msg = _("Moving Message from %s to %s");
	const char *mv_msgs = _("Moving Messages from %s to %s");
	const char *cp_msg = _("Copying Message from %s to %s");
	const char *cp_msgs = _("Copying Messages from %s to %s");
	const char *msg = NULL;
	CORBA_Environment ev[1];
	BRUTUS_ENTRYID *dest_eid = NULL;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_seq_ENTRYID msg_id_list = {
		._maximum = 0,
		._length = 0,
		._buffer = NULL,
	};

	d(source->full_name);
	d(dest->full_name);

	if (!dest_folder->is_a_folder)
		return;

	if (!uids->len)
		return;

	if (!dest_folder->online || !source_folder->online) {
		camel_exception_set(ex, CAMEL_EXCEPTION_SERVICE_NOT_CONNECTED,
				    _("Not connected - The account is read-only"));
		return;
	}

	if (transferred_uids)
		*transferred_uids = NULL;

	CORBA_exception_init(ev);

	// set public message
	if (1 == uids->len)
		msg = delete_originals ? mv_msg : cp_msg;
	else
		msg = delete_originals ? mv_msgs : cp_msgs;

	camel_operation_start(NULL, (char*)msg, source->name, dest->name);

	// sanity checks
	BRUTUS_PRIV_READ_LOCK(priv, folder_tree_rw_mutex);
	dest_eid = (BRUTUS_ENTRYID*)g_hash_table_lookup(priv->entryid_from_full_name_hash, source->full_name);
	if (!dest_eid) {
		BRUTUS_PRIV_READ_UNLOCK(priv, folder_tree_rw_mutex);
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("Source folder not served by Brutus provider"));
		goto out;
	}
	dest_eid = (BRUTUS_ENTRYID*)g_hash_table_lookup(priv->entryid_from_full_name_hash, dest->full_name);
	if (!dest_eid) {
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("Destination folder not served by Brutus provider"));
		goto out;
	}

	// convert uid array to seq_ENTRYID
	if (!brutus_uid_array_to_seq_entryid(uids, &msg_id_list)) {
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("Message IDs are garbled"));
		goto out;
	}

	br = BRUTUS_IMAPIFolder_CopyMessages(source_folder->mapi_folder,
					     &msg_id_list,
					     "",
					     dest_eid,
					     CORBA_OBJECT_NIL,
					     delete_originals ? BRUTUS_BRUTUS_MESSAGE_MOVE : 0,
					     ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		camel_exception_set(ex, CAMEL_EXCEPTION_SERVICE_UNAVAILABLE,
				    _("Could not transfer messages: Communication error"));
	}

out:
	camel_operation_end(NULL);
	BRUTUS_PRIV_READ_UNLOCK(priv, folder_tree_rw_mutex);

	switch (br) {
	case BRUTUS_BRUTUS_MAPI_W_PARTIAL_COMPLETION :
		d(brutus_bresult_to_str(br));
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("Could not transfer some messages"));
	case BRUTUS_BRUTUS_S_OK :
		brutus_update_folder_summary(source, FALSE, ex);
		brutus_update_folder_summary(dest, FALSE, ex);

		break;
	default :
		d(brutus_bresult_to_str(br));
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("Could not transfer any messages"));
	}

	CORBA_exception_free(ev);

	brutus_free_seq_entryid_content(&msg_id_list);

	return;
}

static void
camel_brutus_folder_class_init(CamelBrutusFolderClass *camel_brutus_folder_class)
{
	CamelFolderClass *camel_folder_class = CAMEL_FOLDER_CLASS(camel_brutus_folder_class);

	parent_class = CAMEL_OFFLINE_FOLDER_CLASS(camel_type_get_global_classfuncs(camel_offline_folder_get_type ()));

	((CamelObjectClass *) camel_folder_class)->getv = brutus_folder_getv;

	camel_folder_class->refresh_info = camel_brutus_folder_refresh_info;
	camel_folder_class->sync = camel_brutus_folder_sync;
	camel_folder_class->expunge = camel_brutus_folder_expunge;
	camel_folder_class->append_message = camel_brutus_folder_append_message;
	camel_folder_class->get_message = camel_brutus_folder_get_message;
	camel_folder_class->has_search_capability = camel_brutus_folder_has_search_capability;
	camel_folder_class->transfer_messages_to = camel_brutus_folder_transfer_messages_to;
/*	camel_folder_class->search_by_expression = camel_brutus_folder_search_by_expression; */
/*	camel_folder_class->search_by_uids = camel_brutus_folder_search_by_uids; */
/*	camel_folder_class->search_free = camel_brutus_folder_search_free; */
}

static void
camel_brutus_folder_init(gpointer object,
			 gpointer klass)
{
	CamelBrutusFolder *brutus_folder = CAMEL_BRUTUS_FOLDER(object);
	CamelFolder *folder = CAMEL_FOLDER(object);

	folder->description = NULL;
	folder->summary = NULL;
	folder->permanent_flags =
		CAMEL_MESSAGE_ANSWERED
		| CAMEL_MESSAGE_ANSWERED_ALL
		| CAMEL_MESSAGE_DELETED
		| CAMEL_MESSAGE_DRAFT
		| CAMEL_MESSAGE_FLAGGED
		| CAMEL_MESSAGE_SEEN
		| CAMEL_MESSAGE_JUNK
		| CAMEL_MESSAGE_USER;

	folder->folder_flags =
		CAMEL_FOLDER_HAS_SUMMARY_CAPABILITY
		/* | CAMEL_FOLDER_HAS_SEARCH_CAPABILITY */;

	brutus_folder->is_a_folder = FALSE;
	brutus_folder->online = FALSE;
	brutus_folder->mapi_session = CORBA_OBJECT_NIL;
	brutus_folder->mapi_folder = CORBA_OBJECT_NIL;
	brutus_folder->search = NULL;
	brutus_folder->summary_mutex = g_mutex_new();
	brutus_folder->cache_mutex = g_mutex_new();
	brutus_folder->finalize_mutex = g_mutex_new();
	brutus_folder->offline_caching_thread = NULL;
	brutus_folder->msg_cache = NULL;
	brutus_folder->brutus_server = NULL;

	memset((void*)brutus_folder->session_key, '\0', sizeof(brutus_folder->session_key));
}

static void
camel_brutus_folder_finalize(CamelObject *object)
{
	CORBA_boolean cb = CORBA_TRUE;
	CamelFolder *folder = CAMEL_FOLDER(object);
	CamelBrutusSummary *cbs = CAMEL_BRUTUS_SUMMARY(folder->summary);
	CamelBrutusFolder *brutus_folder = CAMEL_BRUTUS_FOLDER(object);
	CORBA_Environment ev[1];

	d("");

	// make the caching thread terminate
	if (brutus_folder->offline_caching_thread) {
		BRUTUS_FOLDER_LOCK(brutus_folder, finalize_mutex);
		g_thread_join(brutus_folder->offline_caching_thread);
		brutus_folder->offline_caching_thread = NULL;
		BRUTUS_FOLDER_UNLOCK(brutus_folder, finalize_mutex);
	}

	if (((CamelOfflineFolder*)folder)->sync_offline)
		cbs->flags |= BRUTUS_FOLDER_SYNC_OFFLINE;
	else
		cbs->flags &= ~BRUTUS_FOLDER_SYNC_OFFLINE;
	camel_folder_summary_touch(folder->summary);
	camel_folder_summary_save(folder->summary);

	CORBA_exception_init(ev);

	brutus_release_object(BRUTUS_IMAPIFolder_tc, &brutus_folder->mapi_folder, CORBA_OBJECT_NIL, ev);
	PRINT_IF_ORBIT2_EX(ev);

	cb = CORBA_Object_is_nil(brutus_folder->mapi_session, ev);
	if (!ORBIT2_EX(ev) && !cb)
		CORBA_Object_release(brutus_folder->mapi_session, ev);

	CORBA_exception_free(ev);

	if (brutus_folder->summary_mutex)
		g_mutex_free(brutus_folder->summary_mutex);

	if (brutus_folder->cache_mutex)
		g_mutex_free(brutus_folder->cache_mutex);

	if (brutus_folder->search)
		camel_object_unref(CAMEL_OBJECT(brutus_folder->search));

	if (brutus_folder->msg_cache)
		camel_object_unref(CAMEL_OBJECT(brutus_folder->msg_cache));

	if (brutus_folder->brutus_server)
		g_free(brutus_folder->brutus_server);
}

static int
brutus_folder_getv(CamelObject *object,
		   CamelException *ex,
		   CamelArgGetV *args)
{
	CamelFolder *folder = (CamelFolder*)object;
	int i;
	guint32 tag;
	gboolean unhandled = FALSE;

	for (i = 0; i < args->argc; i++) {
		CamelArgGet *arg = &args->argv[i];

		tag = arg->tag;
		switch (tag & CAMEL_ARG_TAG) {
		case CAMEL_OBJECT_ARG_DESCRIPTION:
			if (!folder->description) {
				CamelURL *uri = ((CamelService*)camel_folder_get_parent_store(folder))->url;
				folder->description = g_strdup_printf("%s@%s:%s", uri->user, uri->host, folder->full_name);
			}
			*arg->ca_str = folder->description;
			break;
		default:
			unhandled = TRUE;
			continue;
		}
		arg->tag = (tag & CAMEL_ARG_TYPE) | CAMEL_ARG_IGNORE;
	}
	if (unhandled)
		return ((CamelObjectClass *)parent_class)->getv(object, ex, args);

	return 0;
}

CamelType
camel_brutus_folder_get_type(void)
{
	static CamelType camel_brutus_folder_type = CAMEL_INVALID_TYPE;

	if (camel_brutus_folder_type == CAMEL_INVALID_TYPE) {
		camel_brutus_folder_type =
			camel_type_register(camel_offline_folder_get_type (),
					    "CamelBrutusFolder",
					    sizeof(CamelBrutusFolder),
					    sizeof(CamelBrutusFolderClass),
					    (CamelObjectClassInitFunc)camel_brutus_folder_class_init,
					    NULL,
					    (CamelObjectInitFunc)camel_brutus_folder_init,
					    (CamelObjectFinalizeFunc)camel_brutus_folder_finalize);
	}

	return camel_brutus_folder_type;
}

CamelFolder*
camel_brutus_folder_new_offline(CamelStore *store,
				const char *folder_name,
				guint32 flags,
				CamelException *ex)
{
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(store);
	CamelBrutusStorePrivate *priv = brutus_store->priv;
	CamelFolder *folder = NULL;
	CamelFolderInfo *fi = NULL;
	CamelBrutusFolder *brutus_folder = NULL;
	CamelBrutusSummary *cbs = NULL;
	gboolean is_a_folder = TRUE;
	int i;
	char *name = NULL;
	char *summary_file_path = NULL;
	char *cache_file_path = NULL;
	digest_t digest = { 0 }; // md5 hash of long term ENTRYID
	digest_str_t digest_str = { '\0' }; // stringified version of digest

	d("");
	if (!folder_name || !strlen(folder_name))
		return NULL;
	d(folder_name);

	folder = CAMEL_FOLDER(camel_object_new(camel_brutus_folder_get_type()));
	brutus_folder = CAMEL_BRUTUS_FOLDER(folder);
	name = strrchr(folder_name, '/');
	if (name)
		name++;
	else
		name = (char*)folder_name;
	camel_folder_construct(folder, store, folder_name, name);

	BRUTUS_PRIV_READ_LOCK(priv, folder_tree_rw_mutex);
	fi = (CamelFolderInfo*)g_hash_table_lookup(priv->fi_from_full_name_hash, folder_name);
	if (fi && (fi->flags & CAMEL_FOLDER_NOSELECT)) // it is not a real IMAPIFolder - most likely a message store
		is_a_folder = FALSE;
	BRUTUS_PRIV_READ_UNLOCK(priv, folder_tree_rw_mutex);
	if (!fi) {
		dfmt("Ghost folder: %s", folder_name);
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("Could not open folder - Ghost?"));
		goto ex;
	}

	brutus_folder->is_a_folder = is_a_folder;
	brutus_folder->online = FALSE;
	brutus_folder->mapi_session = CORBA_OBJECT_NIL;
	brutus_folder->mapi_folder = CORBA_OBJECT_NIL;
	memset((void*)brutus_folder->session_key, '\0', sizeof(brutus_folder->session_key));

	md5_get_digest(folder_name, strlen(folder_name), digest);
	for (i = 0; i < sizeof(digest); i++)
		snprintf(&digest_str[2*i], 3, "%02x", (guint8)digest[i]);

	// construct the paths
	summary_file_path = g_strdup_printf ("%s/%s/%s", priv->storage_path, FOLDER_SUMMARIES, digest_str);
	cache_file_path = g_strdup_printf ("%s/%s/%s", priv->storage_path, FOLDER_MESSAGE_CACHE, digest_str);

	// create summary and message cache
	folder->summary = camel_brutus_summary_new(folder, summary_file_path);
	g_free(summary_file_path);
	brutus_folder->msg_cache = camel_data_cache_new(cache_file_path, 0, ex);
	g_free(cache_file_path);
	camel_data_cache_set_expire_age(brutus_folder->msg_cache, MESSAGE_CACHE_TIMEOUT);

	cbs = CAMEL_BRUTUS_SUMMARY(folder->summary);
	if (cbs->flags & BRUTUS_FOLDER_SYNC_OFFLINE)
		((CamelOfflineFolder*)folder)->sync_offline = 1;
	else
		((CamelOfflineFolder*)folder)->sync_offline = 0;

	goto out;
ex:
	if (folder) {
		camel_object_unref(CAMEL_OBJECT(folder));
		folder = NULL;
	}
out:
	return folder;
}

/*
 * The service must be connected when calling this function
 */
CamelFolder*
camel_brutus_folder_new(CamelStore *store,
			const char *folder_name,
			guint32 flags,
			CamelException *ex)
{
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(store);
	CamelBrutusStorePrivate *priv = brutus_store->priv;
	CamelFolder *folder = NULL;
	CamelFolderInfo *fi = NULL;
	CamelBrutusFolder *brutus_folder = NULL;
	CamelBrutusSummary *cbs = NULL;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_ENTRYID *eid = NULL;
	gboolean created = FALSE;
	gboolean is_a_folder = TRUE;
	int i;
	char *name = NULL;
	char *summary_file_path = NULL;
	char *cache_file_path = NULL;
	guchar digest[16] = { 0 }; // md5 hash of long term ENTRYID
	gchar digest_str[33] = { '\0' }; // stringified version of digest
	BRUTUS_BDEFINE obj_type;
	BRUTUS_IMAPIFolder new_mapi_folder = CORBA_OBJECT_NIL;
	CORBA_Environment ev[1];

	d("");
	if (!folder_name || !strlen(folder_name))
		return NULL;
	d(folder_name);

	CORBA_exception_init(ev);

	// force a rebuild of the folder tree if we have just re-connected
	brutus_rebuild_folder_tree(store, ex);

	folder = CAMEL_FOLDER(camel_object_new(camel_brutus_folder_get_type()));
	brutus_folder = CAMEL_BRUTUS_FOLDER(folder);
	name = strrchr(folder_name, '/');
	if (name)
		name++;
	else
		name = (char*)folder_name;
	camel_folder_construct(folder, store, folder_name, name);
	brutus_folder->brutus_server = g_strdup(priv->brutus_server);

	BRUTUS_PRIV_READ_LOCK(priv, folder_tree_rw_mutex);

	fi = (CamelFolderInfo*)g_hash_table_lookup(priv->fi_from_full_name_hash, folder_name);
	if (fi && (fi->flags & CAMEL_FOLDER_NOSELECT)) // it is not a real IMAPIFolder - most likely a message store
		is_a_folder = FALSE;

	eid = (BRUTUS_ENTRYID*)g_hash_table_lookup(priv->entryid_from_full_name_hash, folder_name);
	if (!eid) { // not in the hash table, create it
		BRUTUS_IMAPIFolder mapi_folder = CORBA_OBJECT_NIL;
		CORBA_string pr_display_name = NULL;
		CORBA_string pr_comment = NULL;
		char *parent_name = NULL;
		char *tmp = NULL;

		BRUTUS_PRIV_READ_UNLOCK(priv, folder_tree_rw_mutex);

		if (!(flags & CAMEL_STORE_FOLDER_CREATE)) // not in the hash table, do not create it
			goto ex;

		pr_display_name = CORBA_string_dup(name);
		pr_comment = CORBA_string_dup("");
		parent_name = g_strdup(folder_name);
		tmp = strrchr(parent_name, '/');
		*tmp = '\0';

		BRUTUS_PRIV_WRITE_LOCK(priv, folder_tree_rw_mutex);

		eid = (BRUTUS_ENTRYID*)g_hash_table_lookup(priv->entryid_from_full_name_hash, parent_name);
		g_free(parent_name);
		if (!eid) {
			camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
					    _("Internal data corruption - Shutdown Evolution with \"evolution --force-shutdown\" in a shell, then restart Evolution"));
			goto cleanup;
		}

		d("");
		br = BRUTUS_IMAPISession_OpenEntry(priv->mapi_session,
						   eid,
						   "",
						   BRUTUS_BRUTUS_MAPI_BEST_ACCESS,
						   &obj_type,
						   &mapi_folder,
						   ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			camel_exception_set(ex, CAMEL_EXCEPTION_SERVICE_UNAVAILABLE,
					    _("Communication error"));
			goto cleanup;
		}
		if (BRUTUS_BRUTUS_MAPI_E_NOT_FOUND == br) {
			camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
					    _("Internal data corruption - Shutdown Evolution with \"evolution --force-shutdown\" in a shell, then restart Evolution"));
			goto cleanup;
		}
		if (BRUTUS_BRUTUS_S_OK != br)
			goto cleanup;

		br = BRUTUS_IMAPIFolder_CreateFolder(mapi_folder,
						     BRUTUS_BRUTUS_FOLDER_GENERIC,
						     pr_display_name,
						     pr_comment,
						     "",
						     BRUTUS_BRUTUS_OPEN_IF_EXISTS, // folder created after Evolution startup
						     &new_mapi_folder,
						     ev);
		if (BRUTUS_BRUTUS_MAPI_E_COLLISION == br) {
			camel_exception_set(ex, CAMEL_EXCEPTION_FOLDER_INVALID_PATH,
					    _("Could not create folder - Identically named sibling folders not allowed"));
			goto cleanup;
		}
		if (BRUTUS_BRUTUS_S_OK == br) {
			created = TRUE;
			br = BRUTUS_IMAPIFolder_SaveChanges(mapi_folder, BRUTUS_BRUTUS_KEEP_OPEN_READWRITE, ev);
			if (BRUTUS_BRUTUS_S_OK == br) {
				brutus_insert_in_folder_tree(store, parent_name, folder_name, new_mapi_folder, ex);
				eid = (BRUTUS_ENTRYID*)g_hash_table_lookup(priv->entryid_from_full_name_hash, folder_name);
			}
		}

	cleanup:
		BRUTUS_PRIV_WRITE_UNLOCK(priv, folder_tree_rw_mutex);

		brutus_release_object(BRUTUS_IMAPIFolder_tc, &mapi_folder, CORBA_OBJECT_NIL, ev);
		PRINT_IF_ORBIT2_EX(ev);

		CORBA_free(pr_display_name);
		CORBA_free(pr_comment);
		if (camel_exception_is_set(ex))
			goto ex;
	} else if (is_a_folder) { // present in hash table, re-opening
		br = BRUTUS_IMAPISession_OpenEntry(priv->mapi_session,
						   eid,
						   "",
						   BRUTUS_BRUTUS_MAPI_BEST_ACCESS,
						   &obj_type,
						   &new_mapi_folder,
						   ev);
		if (ORBIT2_EX(ev)) {
			BRUTUS_PRIV_READ_UNLOCK(priv, folder_tree_rw_mutex);
			PRINT_ORBIT2_EX(ev);
			camel_exception_set(ex, CAMEL_EXCEPTION_SERVICE_UNAVAILABLE,
					    _("Communication error"));
			goto ex;
		}
		if (BRUTUS_BRUTUS_MAPI_E_NOT_FOUND == br) {
			BRUTUS_PRIV_READ_UNLOCK(priv, folder_tree_rw_mutex);
			camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
					    _("Internal data corruption - Shutdown Evolution with \"evolution --force-shutdown\" in a shell, then restart Evolution"));
			goto ex;
		}
		if (BRUTUS_BRUTUS_MAPI_E_NO_ACCESS == br) {
			BRUTUS_PRIV_READ_UNLOCK(priv, folder_tree_rw_mutex);
			camel_exception_set(ex, CAMEL_EXCEPTION_FOLDER_INSUFFICIENT_PERMISSION,
					    _("Could not open folder - Access Denied"));
			goto ex;
		}
	} else { // just create a folder with NIL mapi_folder
		br = BRUTUS_BRUTUS_S_OK;
	}

	if (BRUTUS_BRUTUS_S_OK == br) {
		brutus_folder->is_a_folder = is_a_folder;
		brutus_folder->online = is_a_folder; // mapi_folder is only valid if is_a_folder is TRUE
		brutus_folder->mapi_session = (BRUTUS_IMAPISession)CORBA_Object_duplicate((CORBA_Object)priv->mapi_session, ev);
		brutus_folder->mapi_folder = new_mapi_folder; // explicitly NIL or valid
		memcpy((void*)brutus_folder->session_key, (const void*)priv->session_key, sizeof(brutus_folder->session_key));
	} else {
		BRUTUS_PRIV_READ_UNLOCK(priv, folder_tree_rw_mutex);

		brutus_release_object(BRUTUS_IMAPIFolder_tc, &new_mapi_folder, CORBA_OBJECT_NIL, ev);
		PRINT_IF_ORBIT2_EX(ev);

		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("Could not open folder"));
		goto ex;
	}

	BRUTUS_PRIV_READ_UNLOCK(priv, folder_tree_rw_mutex);

	md5_get_digest(folder_name, strlen(folder_name), digest);
	for (i = 0; i < sizeof(digest); i++)
		snprintf(&digest_str[2*i], 3, "%02x", (guint8)digest[i]);

	// construct the paths
	summary_file_path = g_strdup_printf ("%s/%s/%s", priv->storage_path, FOLDER_SUMMARIES, digest_str);
	cache_file_path = g_strdup_printf ("%s/%s/%s", priv->storage_path, FOLDER_MESSAGE_CACHE, digest_str);

	// create summary and message cache
	folder->summary = camel_brutus_summary_new(folder, summary_file_path);
	g_free(summary_file_path);
	brutus_folder->msg_cache = camel_data_cache_new(cache_file_path, 0, ex);
	g_free(cache_file_path);
	camel_data_cache_set_expire_age(brutus_folder->msg_cache, MESSAGE_CACHE_TIMEOUT);

	cbs = CAMEL_BRUTUS_SUMMARY(folder->summary);
	if (cbs->flags & BRUTUS_FOLDER_SYNC_OFFLINE)
		((CamelOfflineFolder*)folder)->sync_offline = 1;
	else
		((CamelOfflineFolder*)folder)->sync_offline = 0;

	// create caching thread
	brutus_folder->offline_caching_thread = g_thread_create((GThreadFunc)cache_folder_offline_thread_func,
								(gpointer)folder,
								TRUE,
								NULL);
	goto out;
ex:
	if (folder) {
		camel_object_unref(CAMEL_OBJECT(folder));
		folder = NULL;
	}

out:
	if (ORBIT2_EX(ev)) {
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("CORBA exception caught"));

	}
	CORBA_exception_free(ev);

	if (created) {
		BRUTUS_PRIV_READ_LOCK(priv, folder_tree_rw_mutex);
		camel_object_trigger_event(store, "folder_created", (CamelFolderInfo*)g_hash_table_lookup(priv->fi_from_full_name_hash, folder_name));
		BRUTUS_PRIV_READ_UNLOCK(priv, folder_tree_rw_mutex);
	}
	return folder;
}

/*
 * mapi_folder must be released by caller if NULL is returned.
 * The CamelFolder assumes ownership of mapi_folder if non-NULL is returned.
 */
CamelFolder*
camel_brutus_folder_wrap(CamelStore *store,
			 BRUTUS_IMAPIFolder mapi_folder,
			 CamelException *ex)
{
	CamelBrutusStore *brutus_store = CAMEL_BRUTUS_STORE(store);
	CamelBrutusStorePrivate *priv = brutus_store->priv;
	CamelFolder *folder = NULL;
	CamelBrutusFolder *brutus_folder = NULL;
	const CamelFolderInfo *fi = NULL;
	CORBA_boolean cb = CORBA_TRUE;
	BRUTUS_ENTRYID *eid = NULL;
	int i;
	char *eid_str = NULL;
	char *summary_file_path = NULL;
	char *cache_file_path = NULL;
	guchar digest[16] = { 0 }; // md5 hash of long term ENTRYID
	gchar digest_str[33] = { '\0' }; // stringified version of digest
	CORBA_Environment ev[1];

	CORBA_exception_init(ev);

	cb = CORBA_Object_is_nil(mapi_folder, ev);
	if (ORBIT2_EX(ev))
		goto ex;
	if (cb) {
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("Folder is NIL"));
		goto ex;
	}

	eid = brutus_get_entryid((BRUTUS_IMAPIProp)mapi_folder, BRUTUS_BRUTUS_PR_ENTRYID);

	// force rebuild of folder tree if just re-connecting
	brutus_rebuild_folder_tree(store, ex);

	BRUTUS_PRIV_READ_LOCK(priv, folder_tree_rw_mutex);

	fi = g_hash_table_lookup(priv->fi_from_entryid_hash, eid);
	if (!fi) { // the folder must be present in the hash table by requirement
		BRUTUS_PRIV_READ_UNLOCK(priv, folder_tree_rw_mutex);
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("Internal data corruption - Shutdown Evolution with \"evolution --force-shutdown\" in a shell, then restart Evolution"));
		goto ex;
	}

	folder = CAMEL_FOLDER(camel_object_new(camel_brutus_folder_get_type()));
	brutus_folder = CAMEL_BRUTUS_FOLDER(folder);
	camel_folder_construct(folder, store, fi->full_name, fi->name);

	brutus_folder->is_a_folder = TRUE;
	brutus_folder->online = TRUE;
	brutus_folder->mapi_session = (BRUTUS_IMAPISession)CORBA_Object_duplicate((CORBA_Object)priv->mapi_session, ev);
	brutus_folder->mapi_folder = mapi_folder;

	md5_get_digest(fi->full_name, strlen(fi->full_name), digest);
	BRUTUS_PRIV_READ_UNLOCK(priv, folder_tree_rw_mutex);
	free(eid_str);
	for (i = 0; i < sizeof(digest); i++)
		snprintf(&digest_str[2*i], 3, "%02x", (guint8)digest[i]);

	// construct the paths
	summary_file_path = g_strdup_printf ("%s/%s/%s", priv->storage_path, FOLDER_SUMMARIES, digest_str);
	cache_file_path = g_strdup_printf ("%s/%s/%s", priv->storage_path, FOLDER_MESSAGE_CACHE, digest_str);

	// create summary and message cache
	folder->summary = camel_brutus_summary_new(folder, summary_file_path);
	g_free(summary_file_path);
	brutus_folder->msg_cache = camel_data_cache_new(cache_file_path, 0, ex);
	g_free(cache_file_path);
	camel_data_cache_set_expire_age(brutus_folder->msg_cache, MESSAGE_CACHE_TIMEOUT);

	// create caching thread
	brutus_folder->offline_caching_thread = g_thread_create((GThreadFunc)cache_folder_offline_thread_func,
								(gpointer)folder,
								TRUE,
								NULL);
	goto out;
ex:
	if (folder) {
		camel_object_unref(CAMEL_OBJECT(folder));
		folder = NULL;
	}

out:
	CORBA_free(eid);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		camel_exception_set(ex, CAMEL_EXCEPTION_SYSTEM,
				    _("CORBA exception caught"));

	}
	CORBA_exception_free(ev);

	return folder;
}
