/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/*
 *    Misc Brutus specific definitions. 
 *
 *    Authors:
 *        Jules Colding <colding@omesc.com>
 *
 *    Copyright (C) 2005-2007 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *    MA 02111-1307 USA
 */

#ifndef BRUTUS_H
#define BRUTUS_H 1

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib.h>
#include <glib/gmacros.h>
#include <execinfo.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>

G_BEGIN_DECLS

/*************************
 *       Typedefs        *
 ************************/

typedef guchar digest_t[16];    // MD5 digest
typedef gchar digest_str_t[33]; // stringified version of digest


/*************************
 *      Capabilities     *
 ************************/

/* define BRUTUS_DEBUG=yes in the environment if you want debugging output */
#define ENV_DEBUG "BRUTUS_DEBUG"

/* Prepend before all output */
#define DEBUG_PREFIX "\n** BRUTUS "PACKAGE_VERSION

/* Do we support SSLIOP? */
#define BRUTUS_HAVE_SSL (1)
#undef BRUTUS_HAVE_SSL // not tested yet

/* Do we support encrypted passwords? */
#define BRUTUS_HAVE_ENCRYPTED_AUTH (1)
#undef BRUTUS_HAVE_ENCRYPTED_AUTH  // not implemented yet


/*************************
 *       Debug macros    *
 ************************/

#undef ORBIT2_EX
#define ORBIT2_EX(_ev_) ((_ev_) && ((_ev_)->_major != CORBA_NO_EXCEPTION))

#undef PRINT_ORBIT2_EX
#ifdef BRUTUS_DEBUG
#define PRINT_ORBIT2_EX(_ev_)                                                                        \
        G_STMT_START {                                                                               \
                char __buf[26];                                                                      \
                time_t __cur_time;                                                                   \
                struct tm __gmt;                                                                     \
                if (getenv(ENV_DEBUG)) {                                                             \
                        time(&__cur_time);                                                           \
                        gmtime_r(&__cur_time, &__gmt);                                               \
                        asctime_r(&__gmt, __buf);                                                    \
                        g_strchomp(__buf);                                                           \
		        switch ((_ev_)->_major) {                                                    \
		        case CORBA_NO_EXCEPTION :                                                    \
			        g_print(DEBUG_PREFIX" (%s) - No exception.\n", __buf);               \
			        break;                                                               \
		        case CORBA_USER_EXCEPTION :                                                  \
			        g_print(DEBUG_PREFIX" (%s) - %s@%d: User Exception : %s\n", __buf,   \
                                        __FILE__, __LINE__, CORBA_exception_id ((_ev_)));            \
			        break;                                                               \
		        case CORBA_SYSTEM_EXCEPTION :                                                \
			        g_print(DEBUG_PREFIX" (%s) - %s@%d: System Exception : %s\n", __buf, \
                                        __FILE__, __LINE__, CORBA_exception_id ((_ev_)));            \
			        break;                                                               \
		        default :                                                                    \
			        break;                                                               \
		        }                                                                            \
                        fflush(NULL);                                                                \
		}                                                                                    \
	} G_STMT_END
#else
#define PRINT_ORBIT2_EX(_ev_) G_STMT_START { } G_STMT_END
#endif

#undef PRINT_IF_ORBIT2_EX
#ifdef BRUTUS_DEBUG
#define PRINT_IF_ORBIT2_EX(_ev_)                                                                             \
        G_STMT_START {                                                                                       \
                char __buf[26];                                                                              \
                time_t __cur_time;                                                                           \
                struct tm __gmt;                                                                             \
                if ((_ev_) && ((_ev_)->_major != CORBA_NO_EXCEPTION) && (getenv(ENV_DEBUG))) {               \
                        time(&__cur_time);                                                                   \
                        gmtime_r(&__cur_time, &__gmt);                                                       \
                        asctime_r(&__gmt, __buf);                                                            \
                        g_strchomp(__buf);                                                                   \
		        switch ((_ev_)->_major) {                                                            \
		        case CORBA_USER_EXCEPTION :                                                          \
			        g_print(DEBUG_PREFIX" (%s) - %s() in %s@%d: User Exception : %s\n", __buf,   \
                                        __FUNCTION__, __FILE__, __LINE__, CORBA_exception_id ((_ev_)));      \
			        break;                                                                       \
		        case CORBA_SYSTEM_EXCEPTION :                                                        \
			        g_print(DEBUG_PREFIX" (%s) - %s() in %s@%d: System Exception : %s\n", __buf, \
                                        __FUNCTION__, __FILE__, __LINE__, CORBA_exception_id ((_ev_)));      \
			        break;                                                                       \
		        default :                                                                            \
			        break;                                                                       \
		        }                                                                                    \
                        fflush(NULL);                                                                        \
		}                                                                                            \
	} G_STMT_END
#else
#define PRINT_IF_ORBIT2_EX(_ev_) G_STMT_START { } G_STMT_END
#endif

#undef d
#ifdef BRUTUS_DEBUG
#define d(_str_)                                                                                                            \
        G_STMT_START {                                                                                                      \
                char __buf[26];                                                                                             \
                time_t __cur_time;                                                                                          \
                struct tm __gmt;                                                                                            \
                if (getenv(ENV_DEBUG)) {                                                                                    \
                        time(&__cur_time);                                                                                  \
                        gmtime_r(&__cur_time, &__gmt);                                                                      \
                        asctime_r(&__gmt, __buf);                                                                           \
                        g_strchomp(__buf);                                                                                  \
                        g_print(DEBUG_PREFIX" (%s) - %s() in %s@%d: %s\n", __buf, __FUNCTION__, __FILE__, __LINE__, _str_); \
                        fflush(NULL);                                                                                       \
                }                                                                                                           \
        } G_STMT_END
#else
#define d(_str_) G_STMT_START { } G_STMT_END
#endif

#undef dfmt
#ifdef BRUTUS_DEBUG
#define dfmt(_fmt_, _str_)                                                                                                       \
        G_STMT_START {                                                                                                           \
                char __buf[26];                                                                                                  \
                time_t __cur_time;                                                                                               \
                struct tm __gmt;                                                                                                 \
                if (getenv(ENV_DEBUG)) {                                                                                         \
                        time(&__cur_time);                                                                                       \
                        gmtime_r(&__cur_time, &__gmt);                                                                           \
                        asctime_r(&__gmt, __buf);                                                                                \
                        g_strchomp(__buf);                                                                                       \
                        g_print(DEBUG_PREFIX" (%s) - %s() in %s@%d: "_fmt_"\n", __buf, __FUNCTION__, __FILE__, __LINE__, _str_); \
                        fflush(NULL);                                                                                            \
                }                                                                                                                \
        } G_STMT_END
#else
#define dfmt(_fmt_, _str_) G_STMT_START { } G_STMT_END
#endif

#undef dfmt2
#ifdef BRUTUS_DEBUG
#define dfmt2(_fmt_, _str1_, _str2_)                                                                                                      \
        G_STMT_START {                                                                                                                    \
                char __buf[26];                                                                                                           \
                time_t __cur_time;                                                                                                        \
                struct tm __gmt;                                                                                                          \
                if (getenv(ENV_DEBUG)) {                                                                                                  \
                        time(&__cur_time);                                                                                                \
                        gmtime_r(&__cur_time, &__gmt);                                                                                    \
                        asctime_r(&__gmt, __buf);                                                                                         \
                        g_strchomp(__buf);                                                                                                \
                        g_print(DEBUG_PREFIX" (%s) - %s() in %s@%d: "_fmt_"\n", __buf, __FUNCTION__, __FILE__, __LINE__, _str1_, _str2_); \
                        fflush(NULL);                                                                                                     \
                }                                                                                                                         \
        } G_STMT_END
#else
#define dfmt2(_fmt_, _str1_, _str2_) G_STMT_START { } G_STMT_END
#endif

#undef dbt
#ifdef BRUTUS_DEBUG
#define dbt()                                                                                                                                        \
        G_STMT_START {                                                                                                                               \
                char __buf[26];                                                                                                                      \
                time_t __cur_time;                                                                                                                   \
                struct tm __gmt;                                                                                                                     \
                if (getenv(ENV_DEBUG)) {                                                                                                             \
                        void *array[20];                                                                                                             \
                        size_t size;                                                                                                                 \
                        size_t i;                                                                                                                    \
                        char **strings;                                                                                                              \
                                                                                                                                                     \
                        time(&__cur_time);                                                                                                           \
                        gmtime_r(&__cur_time, &__gmt);                                                                                               \
                        asctime_r(&__gmt, __buf);                                                                                                    \
                        g_strchomp(__buf);                                                                                                           \
                                                                                                                                                     \
                        size = backtrace(array, 20);                                                                                                 \
                        strings = backtrace_symbols(array, size);                                                                                    \
                        g_print(DEBUG_PREFIX" (%s) - %s() in %s@%d: %s\n\n", __buf, __FUNCTION__, __FILE__, __LINE__, "Printing stack backtrace.."); \
                        for (i = 0; i < size; i++)                                                                                                   \
                                g_print("\t%s\n", strings[i]);                                                                                       \
                        free(strings);                                                                                                               \
                        fflush(NULL);                                                                                                                \
                }                                                                                                                                    \
        } G_STMT_END
#else
#define dbt() G_STMT_START { } G_STMT_END
#endif


/*************************
 *      Utility macros   *
 ************************/

#undef BRUTUS_NULL_IF_EMPTY_STRING
#define BRUTUS_NULL_IF_EMPTY_STRING(_str_) G_STMT_START { if (_str_ && !strlen(_str_)) _str_ = NULL; } G_STMT_END


#undef BRUTUS_CHECK_ALLOC
#define BRUTUS_CHECK_ALLOC(_mem_, _ex_)                                                              \
        G_STMT_START {                                                                               \
	        if (!_mem_) {                                                                        \
		        d("No memory");                                                              \
		        if (_ex_) camel_exception_set(_ex_, CAMEL_EXCEPTION_SYSTEM, _("No memory")); \
		        goto out;                                                                    \
	        }                                                                                    \
	} G_STMT_END


/*************************
 *      GTimer macros    *
 ************************/

#undef BRUTUS_TIMER_NEW
#ifdef BRUTUS_DEBUG
#define BRUTUS_TIMER_NEW() GTimer *_brutus_timer_ = NULL; gdouble _elapsed_time_ = 0.0;
#else
#define BRUTUS_TIMER_NEW() G_STMT_START { } G_STMT_END
#endif

#undef BRUTUS_TIMER_INIT
#ifdef BRUTUS_DEBUG
#define BRUTUS_TIMER_INIT()                             \
        G_STMT_START {				        \
		if (getenv(ENV_DEBUG)) {	        \
			_brutus_timer_ = g_timer_new(); \
			g_timer_stop(_brutus_timer_);	\
		}                                       \
	} G_STMT_END
#else
#define BRUTUS_TIMER_INIT() G_STMT_START { } G_STMT_END
#endif

#undef BRUTUS_TIMER_FREE
#ifdef BRUTUS_DEBUG
#define BRUTUS_TIMER_FREE()                              \
        G_STMT_START {				         \
		if (getenv(ENV_DEBUG)) {	         \
			g_timer_destroy(_brutus_timer_); \
			_elapsed_time_ = 0.0;            \
		}                                        \
	} G_STMT_END
#else
#define BRUTUS_TIMER_FREE() G_STMT_START { } G_STMT_END
#endif

#undef BRUTUS_TIMER_START
#ifdef BRUTUS_DEBUG
#define BRUTUS_TIMER_START()                           \
        G_STMT_START {                                 \
                if (getenv(ENV_DEBUG)) {               \
                        g_timer_start(_brutus_timer_); \
                        fflush(NULL);                  \
                }                                      \
        } G_STMT_END
#else
#define BRUTUS_TIMER_START() G_STMT_START { } G_STMT_END
#endif

#undef BRUTUS_TIMER_PRINT_ELAPSED
#ifdef BRUTUS_DEBUG
#define BRUTUS_TIMER_PRINT_ELAPSED()                                                                                                          \
        G_STMT_START {                                                                                                                        \
                if (getenv(ENV_DEBUG)) {                                                                                                      \
                        g_timer_stop(_brutus_timer_);                                                                                         \
                        _elapsed_time_ = g_timer_elapsed(_brutus_timer_, NULL);                                                               \
                        g_print(DEBUG_PREFIX" %s() in %s@%d: Elapsed time = %f seconds\n", __FUNCTION__, __FILE__, __LINE__, _elapsed_time_); \
                        fflush(NULL);                                                                                                         \
                }                                                                                                                             \
        } G_STMT_END
#else
#define BRUTUS_TIMER_PRINT_ELAPSED() G_STMT_START { } G_STMT_END
#endif


G_END_DECLS

#endif /* BRUTUS_H */
