/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/*
 *    Implementation file for Brutus MAPI utility functions.
 *    Copyright (C) 2005-2007 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *    MA 02111-1307 USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "brutus.h"
#include "brutus_mapi.h"
#include "brutus_util.h"
#include "brutus_corba.h"
#include "brutus_string.h"
#include "brutus_conv.h"
#include <servant_impl/types_impl.c>

#include <glib/gtypes.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>

static void
brutus_kill__defined_but_not_used__compile_warnings(void)
{
	BRUTUS_FUZZY_TYPE fuzzy_level = BRUTUS_BRUTUS_FL_FULLSTRING;
	BRUTUS_BDEFINE prop_tag = 0;
	BRUTUS_SPropValue prop_value = { 0 };
	BRUTUS_RELOP_TYPE relop = BRUTUS_BRUTUS_RELOP_LT;
	BRUTUS_BMR_TYPE bmr = BRUTUS_BRUTUS_BMR_EQZ;

	// don't ever execute this function!
	g_assert(FALSE);

	brutus_kill__defined_but_not_used__compile_warnings();

	impl_BRUTUS_SRestrictionContainer__create(CORBA_OBJECT_NIL, CORBA_OBJECT_NIL, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_SAndRestriction__create(NULL, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_SOrRestriction__create(NULL, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_SNotRestriction__create(0, NULL, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_SContentRestriction__create(fuzzy_level, prop_tag, prop_value, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_SPropertyRestriction__create(relop, prop_tag, prop_value, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_SComparePropsRestriction__create(relop, prop_tag, prop_tag, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_SBitMaskRestriction__create(bmr, prop_tag, prop_tag, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_SSizeRestriction__create(relop, prop_tag, 0, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_SExistRestriction__create(0, prop_tag, 0, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_SSubRestriction__create(prop_tag, CORBA_OBJECT_NIL, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_SCommentRestriction__create(CORBA_OBJECT_NIL, NULL, CORBA_OBJECT_NIL, NULL);

	impl_BRUTUS_IUnknown__create(CORBA_OBJECT_NIL, NULL);
}

gboolean
brutus_string_to_seq_entryid(const char *uid,
			     BRUTUS_seq_ENTRYID *entryids)
{
	if (!entryids)
		return FALSE;

	if (!uid) {
		entryids->_maximum = 0;
		entryids->_length = 0;
		entryids->_buffer = NULL;
		return TRUE;
	}

	entryids->_maximum = 1;
	entryids->_length = 1;
	entryids->_buffer = CORBA_sequence_BRUTUS_ENTRYID_allocbuf(entryids->_maximum);
	if (!entryids->_buffer) {
		entryids->_maximum = 0;
		entryids->_length = 0;
		return FALSE;
	}
	CORBA_sequence_set_release(entryids, CORBA_TRUE);

	return brutus_string_to_entryid_internal_alloc(uid, &entryids->_buffer[0]);
}

gboolean
brutus_entryid_to_seq_entryid(const BRUTUS_ENTRYID *eid,
			      BRUTUS_seq_ENTRYID *entryids)
{
	unsigned int n;

	if (!entryids)
		return FALSE;

	if (!eid) {
		entryids->_maximum = 0;
		entryids->_length = 0;
		entryids->_buffer = NULL;
		return TRUE;
	}

	entryids->_maximum = 1;
	entryids->_length = 1;
	entryids->_buffer = CORBA_sequence_BRUTUS_ENTRYID_allocbuf(entryids->_maximum);
	if (!entryids->_buffer) {
		entryids->_maximum = 0;
		entryids->_length = 0;
		return FALSE;
	}
	CORBA_sequence_set_release(entryids, CORBA_TRUE);

	for (n = 0; n < 4; n++)
		entryids->_buffer[0].abFlags[n] = eid->abFlags[n];

	entryids->_buffer[0].ab._maximum = eid->ab._maximum;
	entryids->_buffer[0].ab._length = eid->ab._length;
	entryids->_buffer[0].ab._buffer = BRUTUS_seq_octet_allocbuf(entryids->_buffer[0].ab._maximum);
	for (n = 0; n < eid->ab._length; n++)
		entryids->_buffer[0].ab._buffer[n] = eid->ab._buffer[n];

	return TRUE;
}

void
brutus_free_seq_entryid_content(BRUTUS_seq_ENTRYID *entryids)
{
	unsigned int n;

	if (!entryids)
		return;

	for (n = 0; n < entryids->_maximum; n++)
		CORBA_free(entryids->_buffer[n].ab._buffer);
	CORBA_free(entryids->_buffer);

	entryids->_maximum = 0;
	entryids->_length = 0;
	entryids->_buffer = NULL;
}

gboolean
brutus_uid_array_to_seq_entryid(const GPtrArray *uids,
				BRUTUS_seq_ENTRYID *entryids)
{
	unsigned int n;
	gboolean retv = TRUE;

	if (!entryids)
		return FALSE;

	if (!uids) {
		entryids->_maximum = 0;
		entryids->_length = 0;
		entryids->_buffer = NULL;
		return TRUE;
	}
	entryids->_maximum = uids->len;
	entryids->_length = uids->len;
	entryids->_buffer = CORBA_sequence_BRUTUS_ENTRYID_allocbuf(entryids->_maximum);
	if (!entryids->_buffer) {
		entryids->_maximum = 0;
		entryids->_length = 0;
		return FALSE;
	}
	CORBA_sequence_set_release(entryids, CORBA_TRUE);

	for (n = 0; n < uids->len; n++)
		retv = retv && brutus_string_to_entryid_internal_alloc((const char*)g_ptr_array_index(uids, n),
								       &entryids->_buffer[n]);

	return retv;
}

gboolean
brutus_create_mapi_folder(BRUTUS_IMAPIFolder parent_folder,
			  const char *container_class,
			  char *name)
{
	gboolean retv = FALSE;
	char *tmp = NULL;
	CORBA_char *child_folder_name = NULL;
	BRUTUS_IMAPIFolder new_folder = CORBA_OBJECT_NIL;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_SPropProblemArray *mapi_problems = NULL;
	BRUTUS_SPropValue pv = {
		.ulPropTag = BRUTUS_BRUTUS_PR_CONTAINER_CLASS,
		.Value._d = BRUTUS_BRUTUS_PT_STRING8,
		.Value._u.lpszA = NULL,
	};
	const BRUTUS_seq_SPropValue prop_values = {
		._maximum = 1,
		._length = 1,
		._buffer = &pv,
	};
	CORBA_Environment ev[1];

	if (!name || !strlen(name))
		return FALSE;

	CORBA_exception_init(ev);

	child_folder_name = CORBA_string_dup(name);
	tmp = strchr(child_folder_name, '/');
	if (tmp) {
		*tmp = '\0';
		tmp++;
	}

	br = BRUTUS_IMAPIFolder_CreateFolder(parent_folder,
					     BRUTUS_BRUTUS_FOLDER_GENERIC,
					     child_folder_name,
					     "",
					     "",
					     0,
					     &new_folder,
					     ev);
	if (BRUTUS_BRUTUS_MAPI_E_COLLISION == br) {
		d(brutus_bresult_to_str(br));
		goto out;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		goto out;
	}

	if (container_class) {
		pv.Value._u.lpszA = CORBA_string_dup(container_class);

		mapi_problems = BRUTUS_seq_SPropProblem__alloc();
		mapi_problems->_maximum = 0;
		mapi_problems->_length = 0;
		mapi_problems->_buffer = NULL;

		br = BRUTUS_IMAPIFolder_SetProps(new_folder, &prop_values, FALSE, &mapi_problems, ev);
		CORBA_free(mapi_problems);
		CORBA_free(pv.Value._u.lpszA);

		if (BRUTUS_BRUTUS_S_OK != br)
			d(brutus_bresult_to_str(br));
	}
	br = BRUTUS_IMAPIFolder_SaveChanges(new_folder, (tmp ? BRUTUS_BRUTUS_KEEP_OPEN_READWRITE : 0), ev);
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		goto out;
	}
	if (tmp)
		retv = brutus_create_mapi_folder(new_folder, container_class, tmp);
	else
		retv = TRUE;

out:
	brutus_release_object(BRUTUS_IMAPIFolder_tc, &new_folder, CORBA_OBJECT_NIL, ev);
	PRINT_IF_ORBIT2_EX(ev);

	CORBA_exception_free(ev);

	CORBA_free(child_folder_name);

	return retv;
}

BRUTUS_IMAPIFolder
brutus_open_folder_by_name(BRUTUS_IMAPIFolder parent_folder,
			   const char *display_name)
{
	BRUTUS_IMAPIFolder retv = CORBA_OBJECT_NIL;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_ENTRYID *eid = NULL;
	BRUTUS_IMAPITable hierachy = CORBA_OBJECT_NIL;
	BRUTUS_SRowSet *row_set = NULL;
	BRUTUS_BDEFINE obj_type = 0;
	CORBA_unsigned_long r = 0;
	CORBA_long n = 0;
	static BRUTUS_BDEFINE props[2] = {
		BRUTUS_BRUTUS_PR_ENTRYID,
		BRUTUS_BRUTUS_PR_DISPLAY_NAME,
	};
	static BRUTUS_SPropTagArray proptag_array = {
		._maximum = 2,
		._length = 2,
		._buffer = props,
	};
	static BRUTUS_SSortOrderSet sort_order = {
		.cCategories = 0,
		.cExpanded = 0,
		.aSort._maximum = 0,
		.aSort._length = 0,
		.aSort._buffer = NULL,
	};
	CORBA_Environment ev[1];

	CORBA_exception_init(ev);

	// get hierachy table
	br = BRUTUS_IMAPIFolder_GetHierarchyTable(parent_folder, 0, &hierachy, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (BRUTUS_BRUTUS_S_OK != br)
		goto out;

	// get all rows
	br = brutus_QueryAllRows(hierachy,
				 &proptag_array,
				 CORBA_OBJECT_NIL,
				 &sort_order,
				 0,
				 &row_set,
				 ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (BRUTUS_BRUTUS_S_OK != br)
		goto out;

	if (!row_set->_length)
		goto out;

	// locate the correct child
	for (r = 0;  r <= row_set->_length; r++) {
		if (r == row_set->_length)
			goto out; // not found

		n = brutus_get_srow_index(&row_set->_buffer[r], BRUTUS_BRUTUS_PR_DISPLAY_NAME);
		if (-1 == n)
			continue;

		if (!g_ascii_strcasecmp(display_name, row_set->_buffer[r]._buffer[n].Value._u.lpszA)) {
			n = brutus_get_srow_index(&row_set->_buffer[r], BRUTUS_BRUTUS_PR_ENTRYID);
			if (-1 == n)
				goto out;
			break;
		}
	}
	brutus_sbinary_to_entryid(&row_set->_buffer[r]._buffer[n].Value._u.bin, &eid);

	br = BRUTUS_IMAPIFolder_OpenEntry(parent_folder,
					  eid,
					  "",
					  BRUTUS_BRUTUS_MAPI_BEST_ACCESS,
					  &obj_type,
					  &retv,
					  ev);
	PRINT_IF_ORBIT2_EX(ev);
	if (BRUTUS_BRUTUS_S_OK != br)
		d(brutus_bresult_to_str(br));

out:
	brutus_release_object(BRUTUS_IMAPITable_tc, &hierachy, CORBA_OBJECT_NIL, ev);
	PRINT_IF_ORBIT2_EX(ev);

	CORBA_exception_free(ev);
	CORBA_free(row_set);
	CORBA_free(eid);

	return retv;
}


void
brutus_copy_spropvalue(BRUTUS_SPropValue *dest,
		       const BRUTUS_SPropValue *src)
{
	CORBA_unsigned_long n = 0;
	CORBA_Environment ev[1];

	if (!dest || !src)
		return;

	dest->ulPropTag = src->ulPropTag;
	dest->Value._d = src->Value._d;

	switch (src->Value._d) {
	case BRUTUS_BRUTUS_PT_UNSPECIFIED :
		return;
	case BRUTUS_BRUTUS_PT_LONG :
		dest->Value._u.l = src->Value._u.l;
		return;
	case BRUTUS_BRUTUS_PT_NULL :
		dest->Value._u.x = src->Value._u.x;
		return;
	case BRUTUS_BRUTUS_PT_SHORT :
		dest->Value._u.i = src->Value._u.i;
		return;
	case BRUTUS_BRUTUS_PT_USHORT :
		dest->Value._u.ui = src->Value._u.ui;
		return;
	case BRUTUS_BRUTUS_PT_ULONG :
		dest->Value._u.ul = src->Value._u.ul;
		return;
	case BRUTUS_BRUTUS_PT_FLOAT :
		dest->Value._u.flt = src->Value._u.flt;
		return;
	case BRUTUS_BRUTUS_PT_DOUBLE :
		dest->Value._u.dbl = src->Value._u.dbl;
		return;
	case BRUTUS_BRUTUS_PT_CURRENCY :
		dest->Value._u.cur.Lo = src->Value._u.cur.Lo;
		dest->Value._u.cur.Hi = src->Value._u.cur.Hi;
		return;
	case BRUTUS_BRUTUS_PT_APPTIME :
		dest->Value._u.at = src->Value._u.at;
		return;
	case BRUTUS_BRUTUS_PT_ERROR :
		dest->Value._u.err = src->Value._u.err;
		return;
	case BRUTUS_BRUTUS_PT_BOOLEAN :
		dest->Value._u.b = src->Value._u.b;
		return;
	case BRUTUS_BRUTUS_PT_LONGLONG :
		dest->Value._u.li = src->Value._u.li;
		return;
		return;
	case BRUTUS_BRUTUS_PT_STRING8 :
		dest->Value._u.lpszA = CORBA_string_dup(src->Value._u.lpszA);
		return;
	case BRUTUS_BRUTUS_PT_OBJECT :
		CORBA_exception_init(ev);
		dest->Value._u.obj = CORBA_Object_duplicate(src->Value._u.obj, ev);
		CORBA_exception_free(ev);
		return;
	case BRUTUS_BRUTUS_PT_UNICODE :
		dest->Value._u.lpszW = CORBA_string_dup(src->Value._u.lpszW);
		return;
	case BRUTUS_BRUTUS_PT_SYSTIME :
		dest->Value._u.ft.dwLowDateTime = src->Value._u.ft.dwLowDateTime;
		dest->Value._u.ft.dwHighDateTime = src->Value._u.ft.dwHighDateTime;
		return;
	case BRUTUS_BRUTUS_PT_CLSID :
		dest->Value._u.lpguid = CORBA_string_dup(src->Value._u.lpguid);
		return;
	case BRUTUS_BRUTUS_PT_BINARY :
		dest->Value._u.bin._maximum = src->Value._u.bin._maximum;
		dest->Value._u.bin._length = src->Value._u.bin._length;
		dest->Value._u.bin._buffer = CORBA_sequence_CORBA_octet_allocbuf(dest->Value._u.bin._maximum);
		memcpy((void*)dest->Value._u.bin._buffer,
		       (void*)src->Value._u.bin._buffer,
		       sizeof(CORBA_octet)*dest->Value._u.bin._length);
		return;
	case BRUTUS_BRUTUS_PT_MV_SHORT :
		dest->Value._u.MVi._maximum = src->Value._u.MVi._maximum;
		dest->Value._u.MVi._length = src->Value._u.MVi._length;
		dest->Value._u.MVi._buffer = CORBA_sequence_CORBA_short_allocbuf(dest->Value._u.MVi._maximum);
		memcpy((void*)dest->Value._u.MVi._buffer,
		       (void*)src->Value._u.MVi._buffer,
		       sizeof(CORBA_short)*dest->Value._u.MVi._length);
		return;
	case BRUTUS_BRUTUS_PT_MV_LONG :
		dest->Value._u.MVl._maximum = src->Value._u.MVl._maximum;
		dest->Value._u.MVl._length = src->Value._u.MVl._length;
		dest->Value._u.MVl._buffer = CORBA_sequence_CORBA_long_allocbuf(dest->Value._u.MVl._maximum);
		memcpy((void*)dest->Value._u.MVl._buffer,
		       (void*)src->Value._u.MVl._buffer,
		       sizeof(CORBA_long)*dest->Value._u.MVl._length);
		return;
	case BRUTUS_BRUTUS_PT_MV_FLOAT :
		dest->Value._u.MVflt._maximum = src->Value._u.MVflt._maximum;
		dest->Value._u.MVflt._length = src->Value._u.MVflt._length;
		dest->Value._u.MVflt._buffer = CORBA_sequence_CORBA_float_allocbuf(dest->Value._u.MVflt._maximum);
		memcpy((void*)dest->Value._u.MVflt._buffer,
		       (void*)src->Value._u.MVflt._buffer,
		       sizeof(CORBA_float)*dest->Value._u.MVflt._length);
		return;
	case BRUTUS_BRUTUS_PT_MV_DOUBLE :
		dest->Value._u.MVdbl._maximum = src->Value._u.MVdbl._maximum;
		dest->Value._u.MVdbl._length = src->Value._u.MVdbl._length;
		dest->Value._u.MVdbl._buffer = CORBA_sequence_CORBA_double_allocbuf(dest->Value._u.MVdbl._maximum);
		memcpy((void*)dest->Value._u.MVdbl._buffer,
		       (void*)src->Value._u.MVdbl._buffer,
		       sizeof(CORBA_double)*dest->Value._u.MVdbl._length);
		return;
	case BRUTUS_BRUTUS_PT_MV_CURRENCY :
		dest->Value._u.MVcur._maximum = src->Value._u.MVcur._maximum;
		dest->Value._u.MVcur._length = src->Value._u.MVcur._length;
		dest->Value._u.MVcur._buffer = CORBA_sequence_BRUTUS_CURRENCY_allocbuf(dest->Value._u.MVcur._maximum);
		memcpy((void*)dest->Value._u.MVcur._buffer,
		       (void*)src->Value._u.MVcur._buffer,
		       sizeof(BRUTUS_CURRENCY)*dest->Value._u.MVcur._length);
		return;
	case BRUTUS_BRUTUS_PT_MV_APPTIME :
		dest->Value._u.MVat._maximum = src->Value._u.MVat._maximum;
		dest->Value._u.MVat._length = src->Value._u.MVat._length;
		dest->Value._u.MVat._buffer = CORBA_sequence_CORBA_double_allocbuf(dest->Value._u.MVat._maximum);
		memcpy((void*)dest->Value._u.MVat._buffer,
		       (void*)src->Value._u.MVat._buffer,
		       sizeof(CORBA_double)*dest->Value._u.MVat._length);
		return;
	case BRUTUS_BRUTUS_PT_MV_SYSTIME :
		dest->Value._u.MVft._maximum = src->Value._u.MVft._maximum;
		dest->Value._u.MVft._length = src->Value._u.MVft._length;
		dest->Value._u.MVft._buffer = CORBA_sequence_BRUTUS_FILETIME_allocbuf(dest->Value._u.MVft._maximum);
		memcpy((void*)dest->Value._u.MVft._buffer,
		       (void*)src->Value._u.MVft._buffer,
		       sizeof(BRUTUS_FILETIME)*dest->Value._u.MVft._length);
		return;
	case BRUTUS_BRUTUS_PT_MV_STRING8 :
		dest->Value._u.MVszA._maximum = src->Value._u.MVszA._maximum;
		dest->Value._u.MVszA._length = src->Value._u.MVszA._length;
		dest->Value._u.MVszA._buffer = CORBA_sequence_CORBA_string_allocbuf(dest->Value._u.MVszA._maximum);
		for (n = 0; n < dest->Value._u.MVszA._length; n++)
			dest->Value._u.MVszA._buffer[n] = CORBA_string_dup(src->Value._u.MVszA._buffer[n]);
		return;
	case BRUTUS_BRUTUS_PT_MV_BINARY :
		dest->Value._u.MVbin._maximum = src->Value._u.MVbin._maximum;
		dest->Value._u.MVbin._length = src->Value._u.MVbin._length;
		dest->Value._u.MVbin._buffer = BRUTUS_SBinaryArray_allocbuf(dest->Value._u.MVbin._maximum);
		for (n = 0; n < dest->Value._u.MVbin._length; n++) {
			dest->Value._u.MVbin._buffer[n]._maximum = src->Value._u.MVbin._buffer[n]._maximum;
			dest->Value._u.MVbin._buffer[n]._length = src->Value._u.MVbin._buffer[n]._length;
			dest->Value._u.MVbin._buffer[n]._buffer = CORBA_sequence_CORBA_octet_allocbuf(dest->Value._u.MVbin._buffer[n]._maximum);

			memcpy((void*)dest->Value._u.MVbin._buffer[n]._buffer,
			       (void*)src->Value._u.MVbin._buffer[n]._buffer,
			       sizeof(CORBA_octet)*dest->Value._u.MVbin._buffer[n]._length);
		}
		return;
	case BRUTUS_BRUTUS_PT_MV_UNICODE :
		dest->Value._u.MVszW._maximum = src->Value._u.MVszW._maximum;
		dest->Value._u.MVszW._length = src->Value._u.MVszW._length;
		dest->Value._u.MVszW._buffer = CORBA_sequence_CORBA_string_allocbuf(dest->Value._u.MVszW._maximum);
		for (n = 0; n < dest->Value._u.MVszW._length; n++)
			dest->Value._u.MVszW._buffer[n] = CORBA_string_dup(src->Value._u.MVszW._buffer[n]);
		return;
	case BRUTUS_BRUTUS_PT_MV_CLSID :
		dest->Value._u.MVguid._maximum = src->Value._u.MVguid._maximum;
		dest->Value._u.MVguid._length = src->Value._u.MVguid._length;
		dest->Value._u.MVguid._buffer = CORBA_sequence_CORBA_string_allocbuf(dest->Value._u.MVguid._maximum);
		for (n = 0; n < dest->Value._u.MVguid._length; n++)
			dest->Value._u.MVguid._buffer[n] = CORBA_string_dup(src->Value._u.MVguid._buffer[n]);
		return;
	case BRUTUS_BRUTUS_PT_MV_LONGLONG :
		dest->Value._u.MVli._maximum = src->Value._u.MVli._maximum;
		dest->Value._u.MVli._length = src->Value._u.MVli._length;
		dest->Value._u.MVli._buffer = CORBA_sequence_CORBA_long_long_allocbuf(dest->Value._u.MVli._maximum);
		memcpy((void*)dest->Value._u.MVli._buffer,
		       (void*)src->Value._u.MVli._buffer,
		       sizeof(CORBA_long_long)*dest->Value._u.MVli._length);
		return;
	default :
		g_assert(FALSE);
	}
}

void
brutus_free_spropvalue_content(BRUTUS_SPropValue *pv,
			       PortableServer_POA poa)
{
	CORBA_unsigned_long n = 0;
	CORBA_Environment ev[1];

	switch (pv->Value._d) {
	case BRUTUS_BRUTUS_PT_UNSPECIFIED :
	case BRUTUS_BRUTUS_PT_LONG :
	case BRUTUS_BRUTUS_PT_NULL :
	case BRUTUS_BRUTUS_PT_SHORT :
	case BRUTUS_BRUTUS_PT_USHORT :
	case BRUTUS_BRUTUS_PT_ULONG :
	case BRUTUS_BRUTUS_PT_FLOAT :
	case BRUTUS_BRUTUS_PT_DOUBLE :
	case BRUTUS_BRUTUS_PT_CURRENCY :
	case BRUTUS_BRUTUS_PT_APPTIME :
	case BRUTUS_BRUTUS_PT_ERROR :
	case BRUTUS_BRUTUS_PT_BOOLEAN :
	case BRUTUS_BRUTUS_PT_LONGLONG :
	case BRUTUS_BRUTUS_PT_SYSTIME :
		break;
	case BRUTUS_BRUTUS_PT_STRING8 :
		CORBA_free(pv->Value._u.lpszA);
		break;
	case BRUTUS_BRUTUS_PT_OBJECT :
		CORBA_exception_init(ev);
		brutus_release_object(BRUTUS_IUnknown_tc,
				      &pv->Value._u.obj,
				      poa,
				      ev);
		CORBA_exception_free(ev);
		break;
	case BRUTUS_BRUTUS_PT_UNICODE :
		CORBA_free(pv->Value._u.lpszW);
		break;
	case BRUTUS_BRUTUS_PT_CLSID :
		CORBA_free(pv->Value._u.lpguid);
		break;
	case BRUTUS_BRUTUS_PT_BINARY :
		CORBA_free(pv->Value._u.bin._buffer);
		break;
	case BRUTUS_BRUTUS_PT_MV_SHORT :
		CORBA_free(pv->Value._u.MVi._buffer);
		break;
	case BRUTUS_BRUTUS_PT_MV_LONG :
		CORBA_free(pv->Value._u.MVl._buffer);
		break;
	case BRUTUS_BRUTUS_PT_MV_FLOAT :
		CORBA_free(pv->Value._u.MVflt._buffer);
		break;
	case BRUTUS_BRUTUS_PT_MV_DOUBLE :
		CORBA_free(pv->Value._u.MVdbl._buffer);
		break;
	case BRUTUS_BRUTUS_PT_MV_CURRENCY :
		CORBA_free(pv->Value._u.MVcur._buffer);
		break;
	case BRUTUS_BRUTUS_PT_MV_APPTIME :
		CORBA_free(pv->Value._u.MVat._buffer);
		break;
	case BRUTUS_BRUTUS_PT_MV_SYSTIME :
		CORBA_free(pv->Value._u.MVft._buffer);
		break;
	case BRUTUS_BRUTUS_PT_MV_STRING8 :
		for (n = 0; n < pv->Value._u.MVszA._length; n++)
			CORBA_free(pv->Value._u.MVszA._buffer[n]);
		/* Why, oh why, is this so ??? Most likely an ORBit2 bug. Oh, well.. we'll see a leak when it's fixed :-( */
//		CORBA_free(pv->Value._u.MVszA._buffer);
		break;
	case BRUTUS_BRUTUS_PT_MV_BINARY :
		for (n = 0; n < pv->Value._u.MVbin._length; n++)
			CORBA_free(pv->Value._u.MVbin._buffer[n]._buffer);
		CORBA_free(pv->Value._u.MVbin._buffer);
		break;
	case BRUTUS_BRUTUS_PT_MV_UNICODE :
		for (n = 0; n < pv->Value._u.MVszW._length; n++)
			CORBA_free(pv->Value._u.MVszW._buffer[n]);
		CORBA_free(pv->Value._u.MVszW._buffer);
		break;
	case BRUTUS_BRUTUS_PT_MV_CLSID :
		for (n = 0; n < pv->Value._u.MVguid._length; n++)
			CORBA_free(pv->Value._u.MVguid._buffer[n]);
		CORBA_free(pv->Value._u.MVguid._buffer);
		break;
	case BRUTUS_BRUTUS_PT_MV_LONGLONG :
		CORBA_free(pv->Value._u.MVli._buffer);
		break;
	default :
		g_assert(FALSE);
	}

	memset((void*)pv, 0, sizeof(BRUTUS_SPropValue));
	pv->Value._d = BRUTUS_BRUTUS_PT_UNSPECIFIED;

	return;
}

void
brutus_free_seq_spropvalue_content(BRUTUS_seq_SPropValue *prop_values,
				   PortableServer_POA poa)
{
	CORBA_unsigned_long n = 0;

	if (!prop_values->_buffer)
		goto out;

	for (n = 0; n < prop_values->_maximum; n++)
		brutus_free_spropvalue_content(&prop_values->_buffer[n], poa);
	CORBA_free(prop_values->_buffer);
	prop_values->_buffer = NULL;

out:
	prop_values->_maximum = 0;
	prop_values->_length = 0;
}


void
brutus_free_addrlist_content(BRUTUS_ADRLIST *addr_list,
			     PortableServer_POA poa)
{
	CORBA_unsigned_long a = 0;

	for (a = 0; a < addr_list->_maximum; a++)
		brutus_free_seq_spropvalue_content(&addr_list->_buffer[a].rgPropVals, poa);
	CORBA_free(addr_list->_buffer);

	addr_list->_maximum = 0;
	addr_list->_length = 0;
	addr_list->_buffer = NULL;
}

gboolean
brutus_entryid_is_shortterm(const BRUTUS_ENTRYID *Eid)
{
	if (!Eid)
		return FALSE;

	return (Eid->abFlags[0] ? TRUE : FALSE);
}

BRUTUS_SPropValue*
brutus_open_object_get_one_prop(BRUTUS_IMAPISession Session,
				BRUTUS_IMAPIFolder Folder,
				BRUTUS_ENTRYID *Eid,
				const BRUTUS_BDEFINE PropTag)
{
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	CORBA_boolean cb = CORBA_TRUE;
	BRUTUS_SPropValue *pv = NULL;
	BRUTUS_IMAPIProp object = CORBA_OBJECT_NIL;
	BRUTUS_BDEFINE obj_type;
	CORBA_Environment ev[1];

	CORBA_exception_init(ev);

	cb = CORBA_Object_is_nil(Folder, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out_no_obj;
	}
	if (!cb) {
		br = BRUTUS_IMAPIFolder_OpenEntry(Folder,
						  Eid,
						  "",
						  BRUTUS_BRUTUS_MAPI_BEST_ACCESS,
						  &obj_type,
						  &object,
						  ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto open_with_session;
		}
	} else {
	open_with_session:
		br = BRUTUS_IMAPISession_OpenEntry(Session,
						   Eid,
						   "",
						   BRUTUS_BRUTUS_MAPI_BEST_ACCESS,
						   &obj_type,
						   &object,
						   ev);
	}
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out_no_obj;
	}

	br = brutus_GetOneProp(object,
			       PropTag,
			       &pv,
			       ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		CORBA_free(pv);
		pv = NULL;
		goto out;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		CORBA_free(pv);
		pv = NULL;
	}

out:
	BRUTUS_IMAPIProp_Destroy(object, 0, ev);
	PRINT_IF_ORBIT2_EX(ev);

	CORBA_Object_release(object, ev);
	PRINT_IF_ORBIT2_EX(ev);

out_no_obj:
	CORBA_exception_free(ev);

	return pv;
}


BRUTUS_IMessage
brutus_open_message(BRUTUS_IMAPISession Session,
		    BRUTUS_IMAPIFolder Folder,
		    const char *UID)
{
	BRUTUS_IMessage retv = CORBA_OBJECT_NIL;
	CORBA_boolean cb = CORBA_FALSE;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_BDEFINE obj_type;
	CORBA_string iid = CORBA_string_dup("");
	CORBA_Environment ev[1];
	BRUTUS_ENTRYID *eid = brutus_string_to_entryid(UID);
	if (!eid)
		return CORBA_OBJECT_NIL;

	CORBA_exception_init(ev);

	cb = CORBA_Object_is_nil(Folder, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		return CORBA_OBJECT_NIL;
	}
	if (!cb) {
		br = BRUTUS_IMAPIFolder_OpenEntry(Folder,
						  eid,
						  iid,
						  BRUTUS_BRUTUS_MAPI_BEST_ACCESS,
						  &obj_type,
						  &retv,
						  ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto open_with_session;
		}
	} else {
	open_with_session:
		br = BRUTUS_IMAPISession_OpenEntry(Session,
						   eid,
						   iid,
						   BRUTUS_BRUTUS_MAPI_BEST_ACCESS,
						   &obj_type,
						   &retv,
						   ev);
	}
	CORBA_free(eid);
	CORBA_free(iid);

	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		retv = CORBA_OBJECT_NIL;
	}
	CORBA_exception_free(ev);

	return retv;
}

char*
brutus_entryid_to_string(const BRUTUS_ENTRYID *Eid)
{
	char *retv = NULL;
	CORBA_unsigned_long n;

	if (!Eid)
		return NULL;

	if (!Eid->abFlags[0])
		brutus_append_to_string(&retv, "%s", "0x00");
	else
		brutus_append_to_string(&retv, "0x%.2hhX", Eid->abFlags[0]);

	for (n = 1; n < 4; n++) {
		if (!Eid->abFlags[n])
			brutus_append_to_string(&retv, "%s", " 0x00");
		else
			brutus_append_to_string(&retv, " 0x%#.2hhX", Eid->abFlags[n]);
	}

	for (n = 0; n < Eid->ab._length; n++) {
		if (!Eid->ab._buffer[n])
			brutus_append_to_string(&retv, "%s", " 0x00");
		else
			brutus_append_to_string(&retv, " 0x%.2hhX", Eid->ab._buffer[n]);
	}

	return retv;
}

gboolean
brutus_string_to_entryid_internal_alloc(const char *uid,
					BRUTUS_ENTRYID *eid)
{
	CORBA_unsigned_long n;
	size_t count = 0;
	char *p = (char*)uid;
	char *endp = p;

	if (!uid) {
		memset((void*)eid, 0, sizeof(BRUTUS_ENTRYID));
		return TRUE;
	}

	while ('\0' != *endp) {
		count++;
		if ((ULONG_MAX == strtoul(p, &endp, 16)) && (ERANGE == errno))
			return FALSE;
		p = endp;
	}
	count -= 4;
	if (0 >= count)
		return FALSE;

	p = (char*)uid;
	for (n = 0; n < 4; n++) {
		eid->abFlags[n] = (CORBA_octet)strtoul(p, &endp, 16);
		p = endp;
	}

	eid->ab._maximum = count;
	eid->ab._length = count;
	eid->ab._buffer = CORBA_sequence_CORBA_octet_allocbuf(eid->ab._length);
	if (!eid->ab._buffer)
		return FALSE;
	CORBA_sequence_set_release(&eid->ab, CORBA_FALSE);

	for (n = 0; n < count; n++) {
		eid->ab._buffer[n] = (CORBA_octet)strtoul(p, &endp, 16);
		p = endp;
	}

	return TRUE;
}

BRUTUS_ENTRYID*
brutus_string_to_entryid(const char *uid)
{
	CORBA_unsigned_long n;
	BRUTUS_ENTRYID *retv = NULL;
	size_t count = 0;
	char *p = (char*)uid;
	char *endp = p;

	if (!uid)
		return NULL;

	while ('\0' != *endp) {
		count++;
		if ((ULONG_MAX == strtoul(p, &endp, 16)) && (ERANGE == errno))
			return NULL;
		p = endp;
	}
	count -= 4;
	if (0 >= count)
		return NULL;

	retv = BRUTUS_ENTRYID__alloc();
	p = (char*)uid;
	for (n = 0; n < 4; n++) {
		retv->abFlags[n] = (CORBA_octet)strtoul(p, &endp, 16);
		p = endp;
	}

	retv->ab._maximum = count;
	retv->ab._length = count;
	retv->ab._buffer = CORBA_sequence_CORBA_octet_allocbuf(retv->ab._length);
	CORBA_sequence_set_release(&retv->ab, CORBA_TRUE);

	for (n = 0; n < count; n++) {
		retv->ab._buffer[n] = (CORBA_octet)strtoul(p, &endp, 16);
		p = endp;
	}

	return retv;
}

gboolean
brutus_entryid_cmp(gconstpointer a,
		   gconstpointer b)
{
	const BRUTUS_ENTRYID *left = (const BRUTUS_ENTRYID*)a;
	const BRUTUS_ENTRYID *right = (const BRUTUS_ENTRYID*)b;
	gboolean retv = TRUE;

	if (!a && !b)
		return TRUE;

	if (a == b)
		return TRUE;

	if (!a || !b)
		return FALSE;

	if (left->ab._length != right->ab._length)
		return FALSE;

	retv = retv && (memcmp(left->abFlags, right->abFlags, 4*sizeof(CORBA_octet)) ? FALSE : TRUE);
	if (FALSE == retv)
		return FALSE;

	retv = retv && (memcmp(left->ab._buffer, right->ab._buffer, left->ab._length*sizeof(CORBA_octet)) ? FALSE : TRUE);
	if (FALSE == retv)
		return FALSE;

	return TRUE;
}


BRUTUS_IMAPIFolder
brutus_get_parent_folder(BRUTUS_IMAPISession Session,
			 BRUTUS_IMAPIProp Object)
{
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_BDEFINE obj_type;
	BRUTUS_IMAPIFolder parent_folder = CORBA_OBJECT_NIL;
	CORBA_string iid = CORBA_string_dup("");
	CORBA_Environment ev[1];
	BRUTUS_ENTRYID *eid = brutus_get_entryid(Object, BRUTUS_BRUTUS_PR_PARENT_ENTRYID);
	if (!eid)
		return CORBA_OBJECT_NIL;

	CORBA_exception_init(ev);
	br = BRUTUS_IMAPISession_OpenEntry(Session,
					   eid,
					   iid,
					   BRUTUS_BRUTUS_MAPI_BEST_ACCESS,
					   &obj_type,
					   &parent_folder,
					   ev);
	PRINT_IF_ORBIT2_EX(ev);
	CORBA_exception_free(ev);

	CORBA_free(eid);
	CORBA_free(iid);

	return parent_folder;
}


BRUTUS_ENTRYID*
brutus_entryid_dup(const BRUTUS_ENTRYID *orig)
{
	CORBA_unsigned_long n;
	BRUTUS_ENTRYID *retv = BRUTUS_ENTRYID__alloc();

	for (n = 0; n < 4; n++)
		retv->abFlags[n] = orig->abFlags[n];

	retv->ab._maximum = orig->ab._maximum;
	retv->ab._length = orig->ab._length;
	retv->ab._buffer = CORBA_sequence_CORBA_octet_allocbuf(orig->ab._length);
	CORBA_sequence_set_release(&retv->ab, CORBA_TRUE);

	for (n = 0; n < orig->ab._length; n++)
		retv->ab._buffer[n] = orig->ab._buffer[n];

	return retv;
}

void
brutus_sbinary_to_entryid(const BRUTUS_SBinary *Bin,
			  BRUTUS_ENTRYID **Eid)
{
	CORBA_unsigned_long n;
	CORBA_unsigned_long len = Bin->_length;

	*Eid = BRUTUS_ENTRYID__alloc();
	if (!Eid)
		return;

	for (n = 0; n < 4; n++)
		(*Eid)->abFlags[n] = Bin->_buffer[n];

	(*Eid)->ab._maximum = len - 4;
	(*Eid)->ab._length = (*Eid)->ab._maximum;
	(*Eid)->ab._buffer = CORBA_sequence_CORBA_octet_allocbuf((*Eid)->ab._length);
	CORBA_sequence_set_release(&(*Eid)->ab, CORBA_TRUE);

	for (n = 4; n < len; n++)
		(*Eid)->ab._buffer[n-4] = Bin->_buffer[n];
}

char*
brutus_get_property_as_string(BRUTUS_IMAPIProp Object,
			      const BRUTUS_BDEFINE PropTag)
{
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_SPropValue *pv = NULL;
	char *retv = NULL;
	CORBA_Environment ev[1];
	CORBA_exception_init(ev);

	br = brutus_GetOneProp(Object,
			       PropTag,
			       &pv,
			       ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto exit;
	}
	if (BRUTUS_BRUTUS_S_OK != br)
		goto exit;

	retv = brutus_print_spropvalue(pv);

exit:
	CORBA_free(pv);
	CORBA_exception_free(ev);

	return retv;
}

char*
brutus_get_guid_str(BRUTUS_IMAPIProp Object,
		    const BRUTUS_BDEFINE PropTag)
{
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	CORBA_unsigned_long n = 0;
	BRUTUS_SPropValue *pv = NULL;
	char *retv = NULL;
	CORBA_Environment ev[1];
	CORBA_exception_init(ev);

	br = brutus_GetOneProp(Object,
			       PropTag,
			       &pv,
			       ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto exit;
	}
	if (BRUTUS_BRUTUS_S_OK != br)
		goto exit;

	if (BRUTUS_BRUTUS_PT_BINARY != pv->Value._d)
		goto exit;

	if (pv->Value._u.bin._length) {
		// first long
		if (!pv->Value._u.bin._buffer[0])
			retv = strdup("0x00");
		else
			retv = brutus_to_string("0x%.2hhX", pv->Value._u.bin._buffer[0]);
		for (n = 1; n < 4; n++) {
			if (!pv->Value._u.bin._buffer[n])
				brutus_append_to_string(&retv, "%s", "00");
			else
				brutus_append_to_string(&retv, "%.2hhX", pv->Value._u.bin._buffer[n]);
		}

		// first short
		if (!pv->Value._u.bin._buffer[4])
			brutus_append_to_string(&retv, "%s", " 0x00");
		else
			brutus_append_to_string(&retv, " 0x%.2hhX", pv->Value._u.bin._buffer[4]);
		if (!pv->Value._u.bin._buffer[5])
			brutus_append_to_string(&retv, "%s", "00");
		else
			brutus_append_to_string(&retv, "%.2hhX", pv->Value._u.bin._buffer[5]);

		// second short
		if (!pv->Value._u.bin._buffer[6])
			brutus_append_to_string(&retv, "%s", " 0x00");
		else
			brutus_append_to_string(&retv, " 0x%.2hhX", pv->Value._u.bin._buffer[6]);
		if (!pv->Value._u.bin._buffer[7])
			brutus_append_to_string(&retv, "%s", "00");
		else
			brutus_append_to_string(&retv, "%.2hhX", pv->Value._u.bin._buffer[7]);

		// byte array
		for (n = 8; n < 16; n++) {
			if (!pv->Value._u.bin._buffer[n])
				brutus_append_to_string(&retv, " %s", "0x00");
			else
				brutus_append_to_string(&retv, " 0x%.2hhX", pv->Value._u.bin._buffer[n]);
		}
	}

exit:
	CORBA_free(pv);
	CORBA_exception_free(ev);

	return retv;
}

BRUTUS_ENTRYID*
brutus_get_entryid(BRUTUS_IMAPIProp Object,
		   const BRUTUS_BDEFINE PropTag)
{
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_SPropValue *pv = NULL;
	BRUTUS_ENTRYID *retv = NULL;
	CORBA_Environment ev[1];
	CORBA_exception_init(ev);

	br = brutus_GetOneProp(Object,
			       PropTag,
			       &pv,
			       ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto exit;
	}
	if (BRUTUS_BRUTUS_S_OK != br)
		goto exit;

	brutus_sbinary_to_entryid(&pv->Value._u.bin, &retv);

exit:
	CORBA_free(pv);
	CORBA_exception_free(ev);

	return retv;
}

char*
brutus_print_spropvalue(const BRUTUS_SPropValue *PropValue)
{
	CORBA_unsigned_long n;
	CORBA_unsigned_long i;
	char *retv = NULL;

	switch (PropValue->Value._d) {
	case BRUTUS_BRUTUS_PT_UNSPECIFIED :
		return NULL;
	case BRUTUS_BRUTUS_PT_NULL :
		return brutus_to_string("%ld", (gint32)PropValue->Value._u.x);
	case BRUTUS_BRUTUS_PT_SHORT :
		return brutus_to_string("%hd (0x%.4hX)", PropValue->Value._u.i, PropValue->Value._u.i);
	case BRUTUS_BRUTUS_PT_USHORT :
		return brutus_to_string("0x%.4hX", PropValue->Value._u.ui);
	case BRUTUS_BRUTUS_PT_LONG :
		return brutus_to_string("%ld (0x%.8lX)", (gint32)PropValue->Value._u.l, (gint32)PropValue->Value._u.l);
	case BRUTUS_BRUTUS_PT_ULONG :
		return brutus_to_string("0x%.8lX", (guint32)PropValue->Value._u.ul);
	case BRUTUS_BRUTUS_PT_FLOAT :
		return brutus_to_string("%f", PropValue->Value._u.flt);
	case BRUTUS_BRUTUS_PT_DOUBLE :
		return brutus_to_string("%f", PropValue->Value._u.dbl);
	case BRUTUS_BRUTUS_PT_CURRENCY :
		return brutus_to_string("Lo:%lu Hi:%ld",
					(guint32)PropValue->Value._u.cur.Lo,
					(gint32)PropValue->Value._u.cur.Hi);
	case BRUTUS_BRUTUS_PT_APPTIME :
		return brutus_to_string("%f", PropValue->Value._u.at);
	case BRUTUS_BRUTUS_PT_ERROR :
		return brutus_to_string("0x%.8lX", (guint32)PropValue->Value._u.err);
	case BRUTUS_BRUTUS_PT_BOOLEAN :
		return brutus_to_string("%s", (CORBA_TRUE == PropValue->Value._u.b) ? "true" : "false");
	case BRUTUS_BRUTUS_PT_OBJECT :
		return brutus_to_string("Unknown interface");
	case BRUTUS_BRUTUS_PT_LONGLONG :
		return brutus_to_string("%lld (0x%.8llX)", PropValue->Value._u.li, PropValue->Value._u.li);
	case BRUTUS_BRUTUS_PT_STRING8 :
		return brutus_to_string("%s", PropValue->Value._u.lpszA);
	case BRUTUS_BRUTUS_PT_UNICODE :
		return brutus_to_string("%s", (const char*)PropValue->Value._u.lpszW);
	case BRUTUS_BRUTUS_PT_SYSTIME :
		return brutus_to_string("dwLowDateTime:%lu dwHighDateTime:%lu",
					(guint32)PropValue->Value._u.ft.dwLowDateTime,
					(guint32)PropValue->Value._u.ft.dwHighDateTime);
	case BRUTUS_BRUTUS_PT_CLSID :
		return brutus_to_string("%s", PropValue->Value._u.lpguid);
	case BRUTUS_BRUTUS_PT_BINARY :
		if (PropValue->Value._u.bin._length) {
			if (!PropValue->Value._u.bin._buffer[0])
				retv = strdup("0x00");
			else
				retv = brutus_to_string("0x%.2hhX", PropValue->Value._u.bin._buffer[0]);
			for (n = 1; n < PropValue->Value._u.bin._length; n++) {
				if (!PropValue->Value._u.bin._buffer[n])
					brutus_append_to_string(&retv, "%s", " 0x00");
				else
					brutus_append_to_string(&retv, " 0x%.2hhX", PropValue->Value._u.bin._buffer[n]);
			}
		}
		break;
	case BRUTUS_BRUTUS_PT_MV_SHORT :
		if (PropValue->Value._u.MVi._length) {
			if (!PropValue->Value._u.MVi._buffer[0])
				retv = strdup("0x0000");
			else
				retv = brutus_to_string("%hd", PropValue->Value._u.MVi._buffer[0]);
			for (n = 1; n < PropValue->Value._u.MVi._length; n++) {
				if (!PropValue->Value._u.MVi._buffer[n])
					brutus_append_to_string(&retv, "%s", " 0x0000");
				else
					brutus_append_to_string(&retv, " %hd", PropValue->Value._u.MVi._buffer[n]);
			}
		}
		break;
	case BRUTUS_BRUTUS_PT_MV_LONG :
		if (PropValue->Value._u.MVl._length)
			retv = brutus_to_string("%d", PropValue->Value._u.MVl._buffer[0]);
		for (n = 1; n < PropValue->Value._u.MVl._length; n++)
			brutus_append_to_string(&retv, " %d", PropValue->Value._u.MVl._buffer[n]);
		break;
	case BRUTUS_BRUTUS_PT_MV_FLOAT :
		if (PropValue->Value._u.MVflt._length)
			retv = brutus_to_string("%f", PropValue->Value._u.MVflt._buffer[0]);
		for (n = 1; n < PropValue->Value._u.MVflt._length; n++)
			brutus_append_to_string(&retv, " %f", PropValue->Value._u.MVflt._buffer[n]);
		break;
	case BRUTUS_BRUTUS_PT_MV_DOUBLE :
		if (PropValue->Value._u.MVdbl._length)
			retv = brutus_to_string("%f", PropValue->Value._u.MVdbl._buffer[0]);
		for (n = 1; n < PropValue->Value._u.MVdbl._length; n++)
			brutus_append_to_string(&retv, " %f", PropValue->Value._u.MVdbl._buffer[n]);
		break;
	case BRUTUS_BRUTUS_PT_MV_CURRENCY :
		if (PropValue->Value._u.MVcur._length)
			retv = brutus_to_string("Lo:%lu Hi:%ld",
						(guint32)PropValue->Value._u.MVcur._buffer[0].Lo,
						(gint32)PropValue->Value._u.MVcur._buffer[0].Hi);
		for (n = 1; n < PropValue->Value._u.MVcur._length; n++)
			brutus_append_to_string(&retv, " Lo:%lu Hi:%ld",
						(guint32)PropValue->Value._u.MVcur._buffer[n].Lo,
						(gint32)PropValue->Value._u.MVcur._buffer[n].Hi);
		break;
	case BRUTUS_BRUTUS_PT_MV_APPTIME :
		if (PropValue->Value._u.MVat._length)
			retv = brutus_to_string("%f", PropValue->Value._u.MVat._buffer[0]);
		for (n = 1; n < PropValue->Value._u.MVat._length; n++)
			brutus_append_to_string(&retv, " %f", PropValue->Value._u.MVat._buffer[n]);
		break;
	case BRUTUS_BRUTUS_PT_MV_SYSTIME :
		if (PropValue->Value._u.MVft._length)
			retv = brutus_to_string("dwLowDateTime:%lu dwHighDateTime:%lu",
						(guint32)PropValue->Value._u.MVft._buffer[0].dwLowDateTime,
						(guint32)PropValue->Value._u.MVft._buffer[0].dwHighDateTime);
		for (n = 1; n < PropValue->Value._u.MVft._length; n++)
			brutus_append_to_string(&retv, " dwLowDateTime:%lu dwHighDateTime:%lu",
						(guint32)PropValue->Value._u.MVft._buffer[n].dwLowDateTime,
						(guint32)PropValue->Value._u.MVft._buffer[n].dwHighDateTime);
		break;
	case BRUTUS_BRUTUS_PT_MV_STRING8 : // will use '\r\n ' to comply with RFC2822 (folding)
		if (PropValue->Value._u.MVszA._length)
			retv = brutus_to_string("%s", (const char*)PropValue->Value._u.MVszA._buffer[0]);
		for (n = 1; n < PropValue->Value._u.MVszA._length; n++)
			brutus_append_to_string(&retv, "\r\n %s", (const char*)PropValue->Value._u.MVszA._buffer[n]);
		break;
	case BRUTUS_BRUTUS_PT_MV_BINARY : // will use '\r\n ' to comply with RFC2822 (folding)
		retv = strdup("");
		for (n = 0; n < PropValue->Value._u.MVbin._length; n++) {
			if (PropValue->Value._u.MVbin._buffer[n]._length)
				brutus_append_to_string(&retv, "0x%.2hhX", PropValue->Value._u.MVbin._buffer[n]._buffer[0]);
			for (i = 1; i < PropValue->Value._u.MVbin._buffer[n]._length; i++)
				brutus_append_to_string(&retv, " 0x%.2hhX", PropValue->Value._u.MVbin._buffer[n]._buffer[i]);
			brutus_append_string(&retv, "\r\n ");
		}
		break;
	case BRUTUS_BRUTUS_PT_MV_UNICODE : // will use '\r\n ' to comply with RFC2822 (folding)
		if (PropValue->Value._u.MVszA._length)
			retv = brutus_to_string("%s", (const char*)PropValue->Value._u.MVszW._buffer[0]);
		for (n = 1; n < PropValue->Value._u.MVszW._length; n++)
			brutus_append_to_string(&retv, "\r\n %s", (const char*)PropValue->Value._u.MVszW._buffer[n]);
		break;
	case BRUTUS_BRUTUS_PT_MV_CLSID : // will use '\r\n ' to comply with RFC2822 (folding)
		for (n = 0; n < PropValue->Value._u.MVguid._length; n++)
			brutus_append_to_string(&retv, "%s\r\n ", PropValue->Value._u.MVguid._buffer[n]);
		break;
	case BRUTUS_BRUTUS_PT_MV_LONGLONG :
		if (PropValue->Value._u.MVli._length)
			retv = brutus_to_string("%lld", PropValue->Value._u.MVli._buffer[0]);
		for (n = 1; n < PropValue->Value._u.MVl._length; n++)
			brutus_append_to_string(&retv, " %lld", PropValue->Value._u.MVli._buffer[n]);
		break;
	default :
		d("Unknown Property type");
		return NULL;
	}

	return retv;
}

CORBA_long
brutus_get_srow_index(const BRUTUS_SRow *row,
		      const BRUTUS_BDEFINE PropTag)
{
	CORBA_unsigned_long n = 0;

	for (n = 0; n < row->_length; n++) {
		if (PropTag == row->_buffer[n].ulPropTag)
			return n;
	}

	return -1;
}

CORBA_long brutus_get_spropvalue_index(const BRUTUS_seq_SPropValue *props,
				       const BRUTUS_BDEFINE PropTag)
{
	CORBA_unsigned_long n = 0;

	for (n = 0; n < props->_length; n++) {
		if (PropTag == props->_buffer[n].ulPropTag)
			return n;
	}

	return -1;
}

BRUTUS_BRESULT
brutus_QueryAllRows(BRUTUS_IMAPITable ptable,
		    BRUTUS_SPropTagArray *ptags,
		    BRUTUS_SRestrictionContainer pres,
		    BRUTUS_SSortOrderSet *psort,
		    CORBA_long crowsMax,
		    BRUTUS_SRowSet **pprows,
		    CORBA_Environment *ev)
{
	CORBA_boolean cb = CORBA_FALSE;
	CORBA_unsigned_long status;
	CORBA_long rows_sought = 0;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;

	*pprows = NULL;

	cb = CORBA_Object_is_nil(ptable, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		return BRUTUS_BRUTUS_UNKNOWN_ERROR;
	}
	if (cb)
		return BRUTUS_BRUTUS_MAPI_E_INVALID_PARAMETER;

	// set the cursor at the start of the table
	br = BRUTUS_IMAPITable_SeekRow(ptable, BRUTUS_BRUTUS_BOOKMARK_BEGINNING, 0, &rows_sought, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		return BRUTUS_BRUTUS_UNKNOWN_ERROR;
	}
	if ((BRUTUS_BRUTUS_S_OK != br) && (BRUTUS_BRUTUS_MAPI_W_POSITION_CHANGED != br))
		return br;

	if (ptags->_length) {
		br = BRUTUS_IMAPITable_SetColumns(ptable, ptags, 0 /* BRUTUS_BRUTUS_TBL_BATCH */, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			return BRUTUS_BRUTUS_UNKNOWN_ERROR;
		}
		if (BRUTUS_BRUTUS_S_OK != br)
			return br;
	}

	cb = CORBA_Object_is_nil(pres, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		return BRUTUS_BRUTUS_UNKNOWN_ERROR;
	}
	if (!cb) {
		br = BRUTUS_IMAPITable_Restrict(ptable, pres, 0 /* BRUTUS_BRUTUS_TBL_BATCH */, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			return BRUTUS_BRUTUS_UNKNOWN_ERROR;
		}
		if (BRUTUS_BRUTUS_S_OK != br)
			return br;
	}

	if (psort->aSort._length) {
		br = BRUTUS_IMAPITable_SortTable(ptable, psort, 0 /* BRUTUS_BRUTUS_TBL_BATCH */, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			return BRUTUS_BRUTUS_UNKNOWN_ERROR;
		}
		if (BRUTUS_BRUTUS_S_OK != br)
			return br;
	}

	if (!crowsMax)
		crowsMax = 0x0fffffff;

	br = BRUTUS_BRUTUS_MAPI_E_BUSY;
	do {
		br = BRUTUS_IMAPITable_QueryRows(ptable, crowsMax, BRUTUS_BRUTUS_TBL_NOADVANCE, pprows, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			return BRUTUS_BRUTUS_UNKNOWN_ERROR;
		}
		if (BRUTUS_BRUTUS_MAPI_E_BUSY == br) {
			br = BRUTUS_IMAPITable_WaitForCompletion(ptable, 0, 1000, &status, ev);
			if (ORBIT2_EX(ev)) {
				PRINT_ORBIT2_EX(ev);
				return BRUTUS_BRUTUS_UNKNOWN_ERROR;
			}
			switch (br) {
			case BRUTUS_BRUTUS_MAPI_E_TIMEOUT :
			case BRUTUS_BRUTUS_MAPI_E_NO_SUPPORT :
				br = BRUTUS_BRUTUS_MAPI_E_BUSY;
				break;
			case BRUTUS_BRUTUS_S_OK :
				break;
			default :
				goto out;
			}
		}
	} while (BRUTUS_BRUTUS_MAPI_E_BUSY == br);

out:
	if (*pprows)
		CORBA_sequence_set_release(*pprows, CORBA_TRUE);
	
	d("");
	return br;
}

BRUTUS_BRESULT
brutus_PrepareTable(BRUTUS_IMAPITable ptable,
		    BRUTUS_SPropTagArray *ptags,
		    BRUTUS_SRestrictionContainer pres,
		    BRUTUS_SSortOrderSet *psort,
		    CORBA_Environment *ev)
{
	CORBA_boolean cb = CORBA_FALSE;
	CORBA_long rows_sought = 0;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;

	cb = CORBA_Object_is_nil(ptable, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		return BRUTUS_BRUTUS_UNKNOWN_ERROR;
	}
	if (cb)
		return BRUTUS_BRUTUS_MAPI_E_INVALID_PARAMETER;

	// set the cursor at the start of the table
	br = BRUTUS_IMAPITable_SeekRow(ptable, BRUTUS_BRUTUS_BOOKMARK_BEGINNING, 0, &rows_sought, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		return BRUTUS_BRUTUS_UNKNOWN_ERROR;
	}
	if ((BRUTUS_BRUTUS_S_OK != br) && (BRUTUS_BRUTUS_MAPI_W_POSITION_CHANGED != br))
		return br;

	if (ptags->_length) {
		br = BRUTUS_IMAPITable_SetColumns(ptable, ptags, 0 /* BRUTUS_BRUTUS_TBL_BATCH */, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			return BRUTUS_BRUTUS_UNKNOWN_ERROR;
		}
		if (BRUTUS_BRUTUS_S_OK != br)
			return br;
	}

	cb = CORBA_Object_is_nil(pres, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		return BRUTUS_BRUTUS_UNKNOWN_ERROR;
	}
	if (!cb) {
		br = BRUTUS_IMAPITable_Restrict(ptable, pres, 0 /* BRUTUS_BRUTUS_TBL_BATCH */, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			return BRUTUS_BRUTUS_UNKNOWN_ERROR;
		}
		if (BRUTUS_BRUTUS_S_OK != br)
			return br;
	}

	if (psort->aSort._length) {
		br = BRUTUS_IMAPITable_SortTable(ptable, psort, 0 /* BRUTUS_BRUTUS_TBL_BATCH */, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			return BRUTUS_BRUTUS_UNKNOWN_ERROR;
		}
		if (BRUTUS_BRUTUS_S_OK != br)
			return br;
	}

	return br;
}

BRUTUS_BRESULT
brutus_QuerySomeRows(BRUTUS_IMAPITable ptable,
		     CORBA_long crowsMax,
		     BRUTUS_SRowSet **pprows,
		     CORBA_Environment *ev)
{
	CORBA_boolean cb = CORBA_FALSE;
	CORBA_unsigned_long status;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;

	*pprows = NULL;

	cb = CORBA_Object_is_nil(ptable, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		return BRUTUS_BRUTUS_UNKNOWN_ERROR;
	}
	if (cb)
		return BRUTUS_BRUTUS_MAPI_E_INVALID_PARAMETER;

	if (!crowsMax)
		crowsMax = 0x0fffffff;

	br = BRUTUS_BRUTUS_MAPI_E_BUSY;
	do {
		br = BRUTUS_IMAPITable_QueryRows(ptable, crowsMax, 0, pprows, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			return BRUTUS_BRUTUS_UNKNOWN_ERROR;
		}
		if (BRUTUS_BRUTUS_MAPI_E_BUSY == br) {
			br = BRUTUS_IMAPITable_WaitForCompletion(ptable, 0, 1000, &status, ev);
			if (ORBIT2_EX(ev)) {
				PRINT_ORBIT2_EX(ev);
				return BRUTUS_BRUTUS_UNKNOWN_ERROR;
			}
			switch (br) {
			case BRUTUS_BRUTUS_MAPI_E_TIMEOUT :
			case BRUTUS_BRUTUS_MAPI_E_NO_SUPPORT :
				br = BRUTUS_BRUTUS_MAPI_E_BUSY;
				break;
			case BRUTUS_BRUTUS_S_OK :
				break;
			default :
				goto out;
			}
		}
	} while (BRUTUS_BRUTUS_MAPI_E_BUSY == br);

out:
	if (*pprows)
		CORBA_sequence_set_release(*pprows, CORBA_TRUE);

	return br;
}

BRUTUS_BRESULT
brutus_QueryRowCount(BRUTUS_IMAPITable ptable,
		     CORBA_unsigned_long *row_count,
		     CORBA_Environment *ev)
{
	CORBA_boolean cb = CORBA_FALSE;
	CORBA_unsigned_long status;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;

	cb = CORBA_Object_is_nil(ptable, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		return BRUTUS_BRUTUS_UNKNOWN_ERROR;
	}
	if (cb)
		return BRUTUS_BRUTUS_MAPI_E_INVALID_PARAMETER;

	br = BRUTUS_BRUTUS_MAPI_E_BUSY;
	do {
		br = BRUTUS_IMAPITable_GetRowCount(ptable, 0, row_count, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			return BRUTUS_BRUTUS_UNKNOWN_ERROR;
		}
		if (BRUTUS_BRUTUS_MAPI_E_BUSY == br) {
			br = BRUTUS_IMAPITable_WaitForCompletion(ptable, 0, 1000, &status, ev);
			if (ORBIT2_EX(ev)) {
				PRINT_ORBIT2_EX(ev);
				return BRUTUS_BRUTUS_UNKNOWN_ERROR;
			}
			switch (br) {
			case BRUTUS_BRUTUS_MAPI_E_TIMEOUT :
			case BRUTUS_BRUTUS_MAPI_E_NO_SUPPORT :
				br = BRUTUS_BRUTUS_MAPI_E_BUSY;
				break;
			case BRUTUS_BRUTUS_S_OK :
				break;
			default :
				goto out;
			}
		}
	} while (BRUTUS_BRUTUS_MAPI_E_BUSY == br);

out:
	return br;
}

BRUTUS_BRESULT
brutus_GetOneProp(BRUTUS_IMAPIProp MapiProp,
		  const BRUTUS_BDEFINE PropTag,
		  BRUTUS_SPropValue **PropValue,
		  CORBA_Environment *ev)
{
	BRUTUS_seq_SPropValue *props = NULL;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	CORBA_boolean cb = CORBA_FALSE;
	BRUTUS_SPropTagArray proptag_array;

	*PropValue = NULL;
	cb = CORBA_Object_is_nil(MapiProp, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		return BRUTUS_BRUTUS_UNKNOWN_ERROR;
	}
	if (cb)
		return BRUTUS_BRUTUS_MAPI_E_INVALID_PARAMETER;

	proptag_array._maximum = 1;
	proptag_array._length = 1;
	proptag_array._buffer = CORBA_sequence_BRUTUS_BDEFINE_allocbuf(proptag_array._length);
	proptag_array._buffer[0] = PropTag;

	br = BRUTUS_IMAPIProp_GetProps(MapiProp,
				       &proptag_array,
				       0,
				       &props,
				       ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		return BRUTUS_BRUTUS_UNKNOWN_ERROR;
	}
	if (BRUTUS_BRUTUS_S_OK == br)
		*PropValue = ORBit_copy_value(&props->_buffer[0], TC_BRUTUS_SPropValue);

	CORBA_free(props);

	return br;
}

/*
 * I do generally prefer to postpone declaration of variables
 * until they are needed, but my use of goto makes this approach
 * sub-optimal here.
 */
BRUTUS_BRESULT
brutus_get_default_message_store(BRUTUS_IMAPISession Session,
				 BRUTUS_IMsgStore *MsgStore,
				 CORBA_Environment *ev)
{
	CORBA_boolean cb = CORBA_FALSE;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_IMAPITable mapi_table = CORBA_OBJECT_NIL;
	CORBA_unsigned_long store_num = 0;
	BRUTUS_SPropTagArray *proptag_array = NULL;
	BRUTUS_SSortOrderSet sort_order;
	BRUTUS_SRowSet *row_set = NULL;
	BRUTUS_ENTRYID *eid = NULL;
	CORBA_string iid = CORBA_string_dup(""); // use default interface (IMsgStore) */

	// sanity check
	cb = CORBA_Object_is_nil(Session, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto exit;
	}
	if (cb)
		goto exit;

	// get message store table
	br = BRUTUS_IMAPISession_GetMsgStoresTable(Session, 0, &mapi_table, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto exit;
	}
	if (BRUTUS_BRUTUS_S_OK != br)
		goto exit;

	// I am using a pointer as it it easier to release the entire
	// array at once than just the _buffer
	proptag_array = BRUTUS_SPropTagArray__alloc();
	proptag_array->_maximum = 2;
	proptag_array->_length = 2;
	proptag_array->_buffer = CORBA_sequence_BRUTUS_BDEFINE_allocbuf(proptag_array->_length);
	proptag_array->_buffer[0] = BRUTUS_BRUTUS_PR_ENTRYID;
	proptag_array->_buffer[1] = BRUTUS_BRUTUS_PR_DEFAULT_STORE;
	CORBA_sequence_set_release(proptag_array, CORBA_TRUE);

	// empty sort order
	sort_order.cCategories = 0;
	sort_order.cExpanded = 0;
	sort_order.aSort._maximum = 0;
	sort_order.aSort._length = 0;
	sort_order.aSort._buffer = NULL;

	// get the single row
	br = brutus_QueryAllRows(mapi_table,
				 proptag_array,
				 CORBA_OBJECT_NIL,
				 &sort_order,
				 0,
				 &row_set,
				 ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto exit;
	}
	if (BRUTUS_BRUTUS_S_OK != br)
		goto exit;

	// get entry id from BLOB
	for (store_num = 0; store_num < row_set->_length; store_num++) {
		if (BRUTUS_BRUTUS_PR_DEFAULT_STORE == row_set->_buffer[store_num]._buffer[1].ulPropTag) {
			if (CORBA_TRUE == row_set->_buffer[store_num]._buffer[1].Value._u.b)
				brutus_sbinary_to_entryid(&row_set->_buffer[store_num]._buffer[0].Value._u.bin, &eid);
		}
	}

	// Open the first returned (default) message store
	br = BRUTUS_IMAPISession_OpenMsgStore(Session,
					      (const BRUTUS_ENTRYID*)eid,
					      iid,
					      (const BRUTUS_BDEFINE)BRUTUS_BRUTUS_MAPI_BEST_ACCESS,
					      MsgStore,
					      ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto exit;
	}
	if (BRUTUS_BRUTUS_S_OK != br)
		goto exit;

exit:
	if (BRUTUS_BRUTUS_S_OK != br) {
		BRUTUS_MAPIERROR *mapi_error = NULL;
		br = BRUTUS_IMAPISession_GetLastError(Session, br, 0, &mapi_error, ev);
		if (ORBIT2_EX(ev))
			PRINT_ORBIT2_EX(ev);
		else if (BRUTUS_BRUTUS_S_OK == br)
			dfmt2("\n\tlpszError     : %s\n\tlpszComponent : %s\n\n", (const char*)mapi_error->lpszError, (const char*)mapi_error->lpszComponent);

		CORBA_free(mapi_error);
	}

	CORBA_free(eid);
	CORBA_free(iid);
	CORBA_free(proptag_array);
	CORBA_free(row_set);

	cb = CORBA_Object_is_nil(mapi_table, ev);
	if (ORBIT2_EX(ev))
		PRINT_ORBIT2_EX(ev);
	else if (!cb) {
		BRUTUS_IMAPITable_Destroy(mapi_table, 0, ev);
		PRINT_IF_ORBIT2_EX(ev);

		CORBA_Object_release(mapi_table, ev);
		PRINT_IF_ORBIT2_EX(ev);
	}

	return br;
}

BRUTUS_BRESULT
brutus_get_root_folder(BRUTUS_IMsgStore MsgStore,
		       BRUTUS_IMAPIFolder *RootFolder,
		       CORBA_Environment *ev)
{
	CORBA_boolean cb = CORBA_FALSE;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_BDEFINE obj_type;
	BRUTUS_IUnknown unk = CORBA_OBJECT_NIL;
	CORBA_string iid = CORBA_string_dup(""); // use default interface (IMAPIFolder)
	BRUTUS_ENTRYID eid = { // open root folder
		.abFlags = { 0 },
		.ab = { 0 }
	};

	// initiate out value asap
	*RootFolder = CORBA_OBJECT_NIL;

	// sanity check
	cb = CORBA_Object_is_nil(MsgStore, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		return BRUTUS_BRUTUS_UNKNOWN_ERROR;
	}
	if (cb)
		return BRUTUS_BRUTUS_MAPI_E_INVALID_PARAMETER;

	do {
		br = BRUTUS_IMsgStore_OpenEntry(MsgStore,
						&eid,
						iid,
						BRUTUS_BRUTUS_MAPI_BEST_ACCESS | BRUTUS_BRUTUS_MAPI_NO_CACHE,
						&obj_type,
						&unk,
						ev);
		if (ORBIT2_EX(ev))
			break;

		if (BRUTUS_BRUTUS_S_OK == br)
			break;

		if (BRUTUS_BRUTUS_MAPI_E_UNKNOWN_FLAGS != br)
			break;

		br = BRUTUS_IMsgStore_OpenEntry(MsgStore,
						&eid,
						iid,
						BRUTUS_BRUTUS_MAPI_BEST_ACCESS,
						&obj_type,
						&unk,
						ev);	
	} while (FALSE);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto exit;
	}
	if (BRUTUS_BRUTUS_S_OK != br)
		goto exit;

	if (BRUTUS_BRUTUS_MAPI_FOLDER == obj_type) {
		*RootFolder = (BRUTUS_IMAPIFolder)CORBA_Object_duplicate(unk, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto exit;
		}
	} else {
		BRUTUS_IUnknown_Destroy(unk, 0, ev);
		PRINT_IF_ORBIT2_EX(ev);
		goto exit;
	}

exit:
	if (BRUTUS_BRUTUS_S_OK != br) {
		BRUTUS_MAPIERROR *mapi_error = NULL;
		br = BRUTUS_IMsgStore_GetLastError(MsgStore, br, 0, &mapi_error, ev);
		if (ORBIT2_EX(ev))
			PRINT_ORBIT2_EX(ev);
		else if (BRUTUS_BRUTUS_S_OK == br)
			dfmt2("\n%s\n%s\n", (const char*)mapi_error->lpszError, (const char*)mapi_error->lpszComponent);

		CORBA_free(mapi_error);
	}

	CORBA_free(iid);

	CORBA_Object_release(unk, ev);
	PRINT_IF_ORBIT2_EX(ev);

	return br;
}

BRUTUS_BRESULT
brutus_get_outlook_folder(BRUTUS_IMsgStore MsgStore,
			  const BRUTUS_BDEFINE PropTag,
			  BRUTUS_IMAPIFolder *Folder,
			  CORBA_Environment *ev)
{
	CORBA_boolean cb = CORBA_FALSE;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_ENTRYID *eid = NULL;
	BRUTUS_SPropValue *pv = NULL;
	CORBA_string iid = CORBA_string_dup(""); // use default interface (IMAPIFolder)
	BRUTUS_BDEFINE obj_type;
	BRUTUS_IUnknown unk = CORBA_OBJECT_NIL;
	BRUTUS_IMAPIFolder root_folder = CORBA_OBJECT_NIL;

	// initiate out value asap
	*Folder = CORBA_OBJECT_NIL;

	// sanity check
	cb = CORBA_Object_is_nil(MsgStore, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		return BRUTUS_BRUTUS_UNKNOWN_ERROR;
	}
	if (cb)
		return BRUTUS_BRUTUS_MAPI_E_INVALID_PARAMETER;

	br = brutus_get_root_folder(MsgStore,
				    &root_folder,
				    ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto exit;
	}
	if (BRUTUS_BRUTUS_S_OK != br)
		goto exit;

	br = brutus_GetOneProp(root_folder,
			       PropTag,
			       &pv,
			       ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto exit;
	}
	if (BRUTUS_BRUTUS_S_OK != br)
		goto exit;

	// get entry id from BLOB
	brutus_sbinary_to_entryid(&pv->Value._u.bin, &eid);

	// open the outlook folder
	br = BRUTUS_IMAPIFolder_OpenEntry(root_folder,
					  eid,
					  iid,
					  BRUTUS_BRUTUS_MAPI_BEST_ACCESS,
					  &obj_type,
					  &unk,
					  ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto exit;
	}
	if (BRUTUS_BRUTUS_S_OK != br)
		goto exit;

	if (BRUTUS_BRUTUS_MAPI_FOLDER == obj_type) {
		*Folder = (BRUTUS_IMAPIFolder)CORBA_Object_duplicate(unk, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto exit;
		}
	} else {
		BRUTUS_IUnknown_Destroy(unk, 0, ev);
		if (ORBIT2_EX(ev)) {
			PRINT_ORBIT2_EX(ev);
			goto exit;
		}
		goto exit;
	}

exit:
	if (BRUTUS_BRUTUS_S_OK != br) {
		BRUTUS_MAPIERROR *mapi_error = NULL;
		br = BRUTUS_IMsgStore_GetLastError(MsgStore, br, 0, &mapi_error, ev);
		if (ORBIT2_EX(ev))
			PRINT_ORBIT2_EX(ev);
		else if (BRUTUS_BRUTUS_S_OK == br)
			dfmt2("\n%s\n%s\n", (const char*)mapi_error->lpszError, (const char*)mapi_error->lpszComponent);

		CORBA_free(mapi_error);
	}

	CORBA_Object_release(unk, ev);
	PRINT_IF_ORBIT2_EX(ev);

	CORBA_free(pv);
	CORBA_free(eid);
	CORBA_free(iid);

	// release server resources
	cb = CORBA_Object_is_nil(root_folder, ev);
	if (ORBIT2_EX(ev))
		PRINT_ORBIT2_EX(ev);
	else if (!cb) {
		BRUTUS_IMAPIFolder_Destroy(root_folder, 0, ev);
		PRINT_IF_ORBIT2_EX(ev);

		CORBA_Object_release(root_folder, ev);
		PRINT_IF_ORBIT2_EX(ev);
	}

	return br;
}

char*
brutus_list_folder(BRUTUS_IMAPIFolder Folder,
		   const BRUTUS_BDEFINE ListBy,
		   CORBA_Environment *ev)
{
	CORBA_boolean cb = CORBA_FALSE;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_SSortOrderSet sort_order;
	BRUTUS_IMAPITable contents_table = CORBA_OBJECT_NIL;
	BRUTUS_SRowSet *row_set = NULL;
	BRUTUS_SPropTagArray *proptag_array = NULL;
	char *tmp = NULL;
	char *retv = NULL;

	// sanity check
	cb = CORBA_Object_is_nil(Folder, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		return NULL;
	}
	if (cb) {
		d("Folder is NIL");
		return NULL;
	}

	br = BRUTUS_IMAPIFolder_GetContentsTable(Folder,
						 0,
						 &contents_table,
						 ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto exit;
	}
	if (BRUTUS_BRUTUS_S_OK != br)
		goto exit;

	// we are only interested in the ListBy property
	proptag_array = BRUTUS_SPropTagArray__alloc();
	proptag_array->_maximum = 1;
	proptag_array->_length = 1;
	proptag_array->_buffer = BRUTUS_SPropTagArray_allocbuf(proptag_array->_length);
	proptag_array->_buffer[0] = ListBy;
	CORBA_sequence_set_release(proptag_array, CORBA_TRUE);

	// empty sort order
	sort_order.cCategories = 0;
	sort_order.cExpanded = 0;
	sort_order.aSort._maximum = 0;
	sort_order.aSort._length = 0;
	sort_order.aSort._buffer = NULL;

	// get the single row
	br = brutus_QueryAllRows(contents_table,
				 proptag_array,
				 CORBA_OBJECT_NIL,
				 &sort_order,
				 0,
				 &row_set,
				 ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto exit;
	}
	if (BRUTUS_BRUTUS_S_OK != br)
		goto exit;

	// write the object values to a string
	if (row_set->_length) {
		CORBA_unsigned_long n;
		for (n=0; n<row_set->_length; n++) {
			brutus_append_to_string(&retv, "        %d: ", n);
			tmp = brutus_print_spropvalue(&row_set->_buffer[n]._buffer[(CORBA_unsigned_long)0]);
			brutus_append_string(&retv, tmp);
			free(tmp);
			tmp = NULL;
			brutus_append_string(&retv, "\n");
		}
	}

exit:
	if (BRUTUS_BRUTUS_S_OK != br) {
		BRUTUS_MAPIERROR *mapi_error = NULL;
		br = BRUTUS_IMAPIFolder_GetLastError(Folder, br, 0, &mapi_error, ev);
		if (ORBIT2_EX(ev))
			PRINT_ORBIT2_EX(ev);
		else if (BRUTUS_BRUTUS_S_OK == br)
			dfmt2("\n%s\n%s\n", (const char*)mapi_error->lpszError, (const char*)mapi_error->lpszComponent);

		CORBA_free(mapi_error);
	}


	CORBA_free(row_set);
	CORBA_free(proptag_array);

	// release server resources
	cb = CORBA_Object_is_nil(contents_table, ev);
	if (ORBIT2_EX(ev))
		PRINT_ORBIT2_EX(ev);
	else if (!cb) {
		BRUTUS_IMAPITable_Destroy(contents_table, 0, ev);
		PRINT_IF_ORBIT2_EX(ev);

		CORBA_Object_release(contents_table, ev);
		PRINT_IF_ORBIT2_EX(ev);
	}

	return retv;
}

char*
brutus_print_properties(BRUTUS_IMAPIProp PropObject,
			CORBA_Environment *ev)
{
	CORBA_boolean cb = CORBA_FALSE;
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_seq_SPropValue *props = NULL;
	BRUTUS_SPropTagArray proptag_array;
	CORBA_unsigned_long n = 0;
	char *tmp = NULL;
	char *retv = NULL;

	// sanity check
	cb = CORBA_Object_is_nil(PropObject, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		return NULL;
	}
	if (cb) {
		d("PropObject is NIL");
		return NULL;
	}

	proptag_array._maximum = 0;
	proptag_array._length = 0;
	proptag_array._buffer = NULL;

	br = BRUTUS_IMAPIProp_GetProps(PropObject,
				       &proptag_array,
				       0,
				       &props,
				       ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		return NULL;
	}
	if ((BRUTUS_BRUTUS_S_OK != br) && (BRUTUS_BRUTUS_MAPI_W_ERRORS_RETURNED != br))
		goto exit;

	for (n=0; n<props->_length; n++) {
		brutus_append_to_string(&retv, "0x%.8lX (%s/%s) :\n",
					(guint32)props->_buffer[n].ulPropTag,
					brutus_proptag_to_str(props->_buffer[n].ulPropTag),
					brutus_proptype_to_str(props->_buffer[n].ulPropTag & 0x0000ffff));
		tmp = brutus_print_spropvalue(&props->_buffer[n]);
		brutus_append_string(&retv, tmp);
		free(tmp);
		tmp = NULL;
		brutus_append_string(&retv, "\n\n");
	}

exit:
	if (BRUTUS_BRUTUS_S_OK != br) {
		BRUTUS_MAPIERROR *mapi_error = NULL;
		br = BRUTUS_IMAPIProp_GetLastError(PropObject, br, 0, &mapi_error, ev);
		if (ORBIT2_EX(ev))
			PRINT_ORBIT2_EX(ev);
		else if (BRUTUS_BRUTUS_S_OK == br)
			dfmt2("\n%s\n%s\n", (const char*)mapi_error->lpszError, (const char*)mapi_error->lpszComponent);

		CORBA_free(mapi_error);
	}

	CORBA_free(props);

	return retv;
}

gboolean 
brutus_server_ping(const char *brutus_server)
{
	int sys_retv = -1;
	char *cmd = NULL;

	if (!brutus_server || !strlen(brutus_server))
		return FALSE;

	cmd = g_strdup_printf("/bin/ping -c 1 %s &>/dev/null", brutus_server);
	if (!cmd)
		return FALSE;
	sys_retv = system(cmd);
	g_free(cmd);
	if (sys_retv)
		return FALSE;

	return TRUE;
}
		    

gboolean
brutus_is_connected(const char *brutus_server,
		    BRUTUS_IMAPISession mapi_session)
{
	gboolean retv = FALSE;
	CORBA_boolean cb = CORBA_TRUE;
	CORBA_Environment ev[1];

	if (!brutus_server || !strlen(brutus_server))
		return FALSE;

#ifndef HAVE_ORBIT2_TIMEOUT
	if (!brutus_server_ping(brutus_server))
		return FALSE;
#endif
	CORBA_exception_init(ev);

	cb = CORBA_Object_is_nil((CORBA_Object)mapi_session, ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (!cb) {
		BRUTUS_IMAPISession_ping(mapi_session, ev);
		if (!ORBIT2_EX(ev))
			retv = TRUE; // we are connected
	}

out:
	CORBA_exception_free(ev);
	return retv;
}


time_t 
brutus_get_server_utc_time(const char *brutus_server,
			   BRUTUS_IMAPISession mapi_session,
			   const char *exchange_server)
{
	BRUTUS_TIME_OF_DAY_INFO exchange_tod;
	char *tmp_exchange_server = NULL;
	time_t utc_time = 0;
	CORBA_boolean cb = CORBA_TRUE;
	CORBA_Environment ev[1];

	if (!brutus_is_connected(brutus_server, mapi_session))
		return 0;

	if (!exchange_server)
		return 0;

	tmp_exchange_server = g_strjoin(NULL, "\\\\", exchange_server, NULL);
	if (!tmp_exchange_server)
		return 0;

	CORBA_exception_init(ev);
	cb = BRUTUS_IMAPISession_GetServerTime(mapi_session, tmp_exchange_server, &exchange_tod, ev);
	g_free(tmp_exchange_server);
	if (!ORBIT2_EX(ev) && cb)
		utc_time = (time_t)exchange_tod.tod_elapsedt;
	CORBA_exception_free(ev);

	return utc_time;
}

BRUTUS_seq_octet*
brutus_read_binary_stream(const BRUTUS_BDEFINE property,
			  BRUTUS_IMAPIProp object)
{
	BRUTUS_BRESULT br = BRUTUS_BRUTUS_UNKNOWN_ERROR;
	BRUTUS_seq_octet *retv = NULL;
	BRUTUS_IStream mapi_stream = CORBA_OBJECT_NIL;
	CORBA_unsigned_long_long stream_size = 0;
	CORBA_Environment ev[1];

	CORBA_exception_init(ev);

	// open stream for text data
	br = BRUTUS_IMAPIProp_OpenProperty(object,
					   property,
					   BRUTUS_BRUTUS_IID_IStream,
					   BRUTUS_BRUTUS_STGM_READ,
					   0,
					   &mapi_stream,
					   ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_proptag_to_str(property));
		d(brutus_bresult_to_str(br));
		goto out;
	}

	// get size of the data
	br = BRUTUS_IStream_GetSize(mapi_stream, 
				    &stream_size, 
				    ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		goto out;
	}

	if (!stream_size)
		goto out;
	if (stream_size > 0xFFFFFFFF) {
		d("Streamed data size too big");
		goto out;
	}

	// read data
	br = brutus_read_from_mapi_stream(mapi_stream, 
					  &retv, 
					  stream_size, 
					  ev);
	if (ORBIT2_EX(ev)) {
		PRINT_ORBIT2_EX(ev);
		goto out;
	}
	if (BRUTUS_BRUTUS_S_OK != br) {
		d(brutus_bresult_to_str(br));
		goto out;
	}

out:
	brutus_release_object(BRUTUS_IStream_tc, &mapi_stream, CORBA_OBJECT_NIL, ev);
	PRINT_IF_ORBIT2_EX(ev);

	CORBA_exception_free(ev);

	return retv;
}

char*
brutus_read_text_stream(const BRUTUS_BDEFINE property,
			BRUTUS_IMAPIProp object)
{
	char *retv = NULL;
	BRUTUS_seq_octet *data = NULL;

	data = brutus_read_binary_stream(property, object);
	if (data && data->_length)
		retv = g_strdup((const char*)data->_buffer);
	CORBA_free(data);

	return retv;
}
