/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/*
 *    Implementation file for Brutus SMTP utility functions.
 *    Copyright (C) 2005-2007 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *    MA 02111-1307 USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <glib.h>
#include <glib/gprintf.h>

#include "brutus.h"
#include "brutus_smtp.h"

char*
brutus_unfold_rfc822_header(const char *rfc822_header)
{
	char *retv = NULL;
	char *dst = NULL;

	retv = (char*)malloc(sizeof(char)*(strlen(rfc822_header) + 1));
	if (!retv)
		return NULL;
	dst = retv;

	while (1) {
		if ('\0' == *rfc822_header)
			break;

		if (('\r' == *rfc822_header) && // using RFC 2822 section 2.2 to exclude testing for '\n'
		    ((' ' == *(rfc822_header + 2)) || ('\t' == *(rfc822_header + 2)))) {
			rfc822_header += 2;
		}
		*dst = *rfc822_header;
		if ('\0' == *dst)
			break;

		dst++;
		rfc822_header++;
	}
	*dst = '\0';

	if (strlen(retv) != strlen(rfc822_header))
		retv = (char*)realloc((void*)retv, strlen(retv) + 1);

	return retv;
}

struct _camel_header_raw*
brutus_get_camel_header_raw(const char *rfc822_header)
{
	size_t size = 0;
	const char *start = NULL;
	const char *end = NULL;
	struct _camel_header_raw *retv = NULL;
	struct _camel_header_raw *retv_pos = NULL;
	struct _camel_header_raw *retv_prev = NULL;

	if (!rfc822_header)
		return NULL;

	retv = g_new0(struct _camel_header_raw, 1);
	retv_pos = retv;
	start = rfc822_header;
	while (1) {
		// delimit the current header name
		end = strchr(start, ':');
		if (!end) {
			if (retv_prev) {
				g_free(retv_pos);
				retv_prev->next = NULL;
			}
			break;
		}

		// copy message header name
		size = sizeof(char) * (end - start);
		retv_pos->name = (gchar*)g_malloc(size + sizeof(char));
		if (!retv_pos->name)
			goto rollback;
		memcpy((void*)retv_pos->name, (void*)start, size);
		retv_pos->name[size] = '\0';

		// delimit the current header value
		start = end + 1; // right after the ':'
		while (g_ascii_isspace(*start))
			start++;
		end = strchr(start, '\r');

		// copy message header value
		size = sizeof(char) * (end - start);
		retv_pos->value = (gchar*)g_malloc(size + sizeof(char));
		if (!retv_pos->value)
			goto rollback;
		memcpy((void*)retv_pos->value, (void*)start, size);
		retv_pos->value[size] = '\0';

		// move start forward
		start = end + 2;

		// the end?
		if ((*start == '\r') // headers end and the message body begins
		    || (!*start))    // headers end but there is no message body
			break;

		// allocate the next field and advance the pointer
		retv_prev = retv_pos;
		retv_pos = g_new0(struct _camel_header_raw, 1);
		retv_prev->next = retv_pos;
		continue;

	rollback:
		g_free(retv_pos->name);
		g_free(retv_pos->value);
		g_free(retv_pos);
		if (retv_prev)
			retv_prev->next = NULL;
		break;
	}

	return retv;
}
